import gzip
import json
import traceback
import SimpleITK as sitk

from pathlib import Path
from loguru import logger
from mircat_v2.configs import read_dbase_config


class Nifti:
    """A class used to represent a nifti file for mircat-v2 - including spacing, metadata, segmentations and others!"""

    def __init__(self, nifti_path: str | Path):
        nifti_path = Path(nifti_path).resolve()
        if not nifti_path.exists():
            message = f"Nifti file not found at {nifti_path}! Please check existence."
            logger.error(message)
            raise FileNotFoundError()
        if ".nii" not in nifti_path.suffixes:
            message = f"Input {nifti_path} not recognized as a nifti file. Please ensure file suffix is *.nii or *.nii.gz"
            logger.error(message)
            raise NotNiftiFileError()
        self.path = nifti_path
        self.parent = nifti_path.parent
        self.name = nifti_path.with_suffix("").stem
        self.seg_folder = self.parent / f"{self.name}_segs"
        self.segmentations = list(self.seg_folder.glob(f"{self.name}*.nii.gz"))
        try:
            self.img = sitk.ReadImage(nifti_path)
        except Exception as e:
            logger.error(
                f"Error loading {nifti_path} with SimpleITK :{e}\n{traceback.format_exc()}"
            )
            raise SimpleITKReadError()
        self.spacing = self.img.GetSpacing()
        self.shape = self.img.GetSize()
        if (nifti_path.parent / "metadata.json").exists():
            logger.debug(
                f"{nifti_path.parent / 'metadata.json'} found. Loading metadata."
            )
            with (nifti_path.parent / "metadata.json").open() as f:
                self.metadata: dict = json.load(f)
        elif (nifti_path.parent / "header_info.json").exists():
            logger.debug(
                f"Old format metadata {nifti_path.parent / 'header_info.json'} found. Loading metadata."
            )
            with (nifti_path.parent / "header_info.json").open() as f:
                self.metadata: dict = json.load(f)
        else:
            logger.debug(
                f"No metadata.json file found next to {nifti_path}. Setting metadata to empty dictionary."
            )
            self.metadata: dict = {}


class SegNifti(Nifti):
    def __init__(self, nifti_path: str | Path):
        super().__init__(nifti_path)
        self.task_files: dict = {}

    def resample_and_save_for_segmentation(
        self, new_spacing, output_path, interpolator_type
    ) -> None:
        if len(new_spacing) == 2:
            # Set the new spacing z-length to be the same as the original
            new_spacing = [*new_spacing, self.spacing[-1]]
        # Resample image - always not a label
        resampled = resample_with_sitk(
            self.img,
            new_spacing=new_spacing,
            is_label=False,
            interpolator_type=interpolator_type,
        )
        sitk.WriteImage(resampled, output_path)
        logger.debug("Successfully resampled and wrote to {}", output_path)


class StatsNifti(Nifti):
    def __init__(self, nifti_path: str | Path, overwrite: bool = False):
        super().__init__(nifti_path)
        # TODO this needs to be updated to account for gzipping option
        stats_file = self.seg_folder / f"{self.name}_stats.json"
        self.stats_file = stats_file

        if stats_file.exists() and not overwrite:
            logger.debug(f"Stats file {stats_file} found. Loading statistics.")
            try:
                with stats_file.open() as f:
                    self.stats = json.load(f)
            except json.JSONDecodeError:
                logger.error(
                    f"Error decoding JSON from {stats_file}. Initializing empty stats."
                )
                self.stats = {}
        else:
            if not overwrite:
                logger.debug(
                    f"No stats file found at {stats_file}. Setting stats to empty dictionary."
                )
            else:
                logger.debug(
                    f"Stats file {stats_file} exists but overwrite is set to True. Initializing empty stats."
                )
            self.stats = {}
        self.add_stats("nifti", str(self.path))
        # We want this initialized regardless of it's it's empty or not
        self.add_stats("metadata", self.metadata)
        if self.metadata != {}:
            logger.debug(f"Merging metadata with stats for {self.name}.")
            self.add_stats("series_uid", self.metadata.get("series_uid", None))
            self.add_stats("study_uid", self.metadata.get("study_uid", None))
        # TODO - currently vertebrae midlines will load regardless of the desired stats resolution - which is not correct.
        # If the stats resolution is different from the nifti resolution, we must remeasure the midlines.
        if self.stats.get("vertebrae") is not None:
            self.set_vertebrae_midlines(self.stats["vertebrae"])
        # Placeholder for database stats
        self.db_stats: dict[str, list[dict]] = {}

    def set_task_to_id_map(self, task_to_id_map: dict) -> None:
        self.task_to_id_map = task_to_id_map
        return self

    def set_id_to_seg_map(self, id_to_seg_map: dict) -> None:
        """Set the mapping of task IDs to task names for this nifti."""
        self.id_to_seg_map = id_to_seg_map
        return self

    def set_vertebrae_midlines(self, vertebrae_stats: dict = None):
        """Set the vertebrae midlines for this nifti."""
        if vertebrae_stats is None:
            if "vertebrae" not in self.stats:
                logger.warning(
                    "No vertebrae stats found in nifti. Cannot set vertebrae midlines."
                )
                return self
            vertebrae_stats = self.stats["vertebrae"]
        vertebrae_midlines = {
            k: v.get("midline")
            for k, v in vertebrae_stats.items()
            if isinstance(v, dict)
        }
        self.vertebrae_midlines = vertebrae_midlines
        return self

    def add_stats(self, key: str, value: any, subdict=None) -> None:
        """Add a key-value pair to the stats dictionary."""
        if key in self.stats:
            logger.debug(
                f"Key {key} already exists in stats for {self.path}. Overwriting value."
            )
        if subdict:
            self.stats[subdict][key] = value
        else:
            self.stats[key] = value
        return self

    def preprocess_for_stats(
        self, new_spacing, image_resampler, label_resampler
    ) -> None:
        logger.debug(
            "Resampling image and segmentations for nifti {} to new spacing {}",
            self.name,
            new_spacing,
        )
        self.img = resample_with_sitk(
            self.img,
            new_spacing=new_spacing,
            is_label=False,
            interpolator_type=image_resampler,
            reorient="LAI",
        )
        self.spacing = self.img.GetSpacing()
        self.shape = self.img.GetSize()
        self.add_stats("stats_resolution", new_spacing)
        new_id_to_seg_map = {}
        for seg_id, seg_path in self.id_to_seg_map.items():
            logger.debug(
                "Resampling segmentation for seg_id {} at path {}", seg_id, seg_path
            )
            label = sitk.ReadImage(seg_path)
            resampled_label = resample_with_sitk(
                label,
                is_label=True,
                interpolator_type=label_resampler,
                reference_image=self.img,
            )
            assert resampled_label.GetSize() == self.img.GetSize(), (
                f"Resampled label size {resampled_label.GetSize()} does not match image size {self.img.GetSize()}"
            )
            new_id_to_seg_map[seg_id] = resampled_label
        self.set_id_to_seg_map(new_id_to_seg_map)
        return self

    def check_for_segmentations(self, task_list, task_map):
        task_to_id_map = {}
        id_to_segmentation_map = {}
        # These are the mircat-v1 output file suffixes for these 3 tasks - we just have to remap to the new values
        old_to_new_map = {"total": "999", "tissues": "481", "body": "299"}
        for task in task_list:
            allowed_segmentations = task_map.get(task, [])
            for seg_id in allowed_segmentations:
                seg_file = f"{self.name}_{seg_id}.nii.gz"
                if seg_file in [seg.name for seg in self.segmentations]:
                    if old_to_new_map.get(seg_id, False):
                        logger.debug(f"Flipping {seg_id} to {old_to_new_map[seg_id]}")
                        seg_id = old_to_new_map[seg_id]
                    task_to_id_map[task] = seg_id
                    id_to_segmentation_map[seg_id] = self.seg_folder / seg_file
                    # We break because the task_map is sorted by preference
                    break
        logger.debug(f"{task_to_id_map=}")
        self.set_task_to_id_map(task_to_id_map)
        self.set_id_to_seg_map(id_to_segmentation_map)
        return self

    def save_json_stats(self, gzip_file: bool = False) -> None:
        """Save the statistics to a JSON file."""
        save_stats = {k: v for k, v in self.stats.items() if "-db" not in k}
        if gzip_file:
            logger.debug(f"Compressing stats file {self.stats_file} with gzip.")
            self.add_stats(
                "output_stats_file",
                str(self.stats_file.with_suffix(".json.gz")),
                "metadata",
            )
            with gzip.open(
                self.stats_file.with_suffix(".json.gz"), "wt", encoding="utf-8"
            ) as f:
                json.dump(save_stats, f, indent=2)
        else:
            self.add_stats("output_stats_file", str(self.stats_file), "metadata")
            with self.stats_file.open("w") as f:
                json.dump(save_stats, f, indent=2)

    def format_stats_for_db(self):
        """Format the stats for database insertion. The output dictionary will have keys matching the table names. This function also has internal functions defined in its scope"""
        stats_tables = [
            "metadata",
            "vol_int",
            "contrast",
            "vertebrae",
            "aorta_metrics",
            "aorta_diameters",
            "tissues_volumetric",
            "tissues_vertebral",
        ]
        formatted_stats: dict[str, list[tuple]] = {}
        nifti_path = str(self.path)
        series_uid = self.stats.get("series_uid", None)
        identifier = {"nifti": nifti_path, "series_uid": series_uid}
        study_uid = self.stats.get("study_uid", None)
        tables = read_dbase_config().get("tables")

        def _format_metadata() -> list[dict]:
            """Format the metadata for database insertion."""
            metadata = self.stats.get("metadata", {})
            if not metadata:
                logger.warning(f"No metadata found in the stats for {self.path}")
                return []
            columns = list(tables["metadata"].keys())
            row_data = {}
            for col in columns:
                value = metadata.get(col)
                if col == "nifti" and col not in metadata:
                    value = nifti_path
                elif col == "series_uid" and col not in metadata:
                    value = series_uid
                elif col == "study_uid" and col not in metadata:
                    value = study_uid
                # There are some columns that have been renamed that we need to check for
                if value is None:
                    if col == "birth_date":
                        value = metadata.get("birthday")
                elif col == "pixel_length_mm":
                    value = metadata.get("length_mm")
                elif col == "pixel_width_mm":
                    value = metadata.get("width_mm")
                elif col == "age":
                    value = metadata.get("age_at_scan")
                row_data[col] = value
            return [row_data]

        def _format_vol_int() -> list[dict]:
            "Format the vol_int data for database insertion"
            vol_int = self.stats.get("vol_int", {})
            if not vol_int:
                logger.warning(f"No vol_int stats found in {self.path}")
                return []
            rows = []
            for structure, stats in vol_int.items():
                rows.append(
                    {
                        **identifier,
                        "structure": structure,
                        "volume_cm3": stats.get("volume_cm3"),
                        "hu_mean": stats.get("hu_mean"),
                        "hu_std_dev": stats.get("hu_std_dev"),
                    }
                )
            return rows

        def _format_contrast() -> list[dict]:
            contrast = self.stats.get("contrast", {})
            if not contrast:
                logger.warning(f"No contrast stats found in {self.path}")
                return []
            return [
                {
                    **identifier,
                    "phase": contrast.get("phase"),
                    "probability": contrast.get("probability"),
                    "pi_time": contrast.get("pi_time"),
                    "pi_time_std": contrast.get("pi_time_std"),
                }
            ]

        def _format_vertebrae() -> list[dict]:
            vertebrae = self.stats.get("vertebrae", {})
            if not vertebrae:
                logger.warning(f"No vertebrae stats found in {self.path}")
                return []
            rows = []
            for vertebra, stats in vertebrae.items():
                rows.append(
                    {
                        **identifier,
                        "vertebra": vertebra,
                        "midline": stats.get("midline"),
                    }
                )
            return rows

        def _format_aorta_metrics() -> list[dict]:
            aorta_metrics = self.stats.get("aorta", {})
            kvp = self.stats.get("metadata", {}).get("kvp", None)
            if kvp is None:
                pass
            elif kvp == 120:
                kvp = 1
            else:
                kvp = 0
            if not aorta_metrics:
                logger.warning(f"No aorta metrics found in {self.path}")
                return []
            rows = []
            for region, stats in aorta_metrics.items():
                rows.append(
                    {
                        **identifier,
                        "region": region,
                        "entire_region": stats.get("entire_region"),
                        "length_mm": stats.get("length_mm"),
                        "tortuosity_index": stats.get("tortuosity_index"),
                        "icm": stats.get("icm"),
                        "n_inflections": stats.get("n_inflections"),
                        "peria_volume_cm3": stats.get("peria_volume_cm3"),
                        "peria_ring_volume_cm3": stats.get("peria_ring_volume_cm3"),
                        "peria_fat_volume_cm3": stats.get("peria_fat_volume_cm3"),
                        "peria_hu_mean": stats.get("peria_hu_mean"),
                        "peria_hu_std": stats.get("peria_hu_std"),
                        "calc_volume_mm3": stats.get("calc_volume_mm3"),
                        "calc_agatston": stats.get("calc_agatston"),
                        "calc_count": stats.get("calc_count"),
                        "is_120_kvp": kvp,
                        "mean_diameter_mm": stats.get("mean", {}).get("diameter"),
                        "mean_roundness": stats.get("mean", {}).get("roundness"),
                        "mean_flatness": stats.get("mean", {}).get("flatness"),
                    }
                )
            return rows

        def _format_aorta_diameters() -> list[dict]:
            aorta_metrics = self.stats.get("aorta", {})
            if not aorta_metrics:
                logger.warning(f"No aorta diameters found in {self.path}")
                return []
            rows = []
            measure_keys = ["max", "min", "mid", "proximal", "distal"]
            for region, stats in aorta_metrics.items():
                if region == "whole":
                    continue  # Skip the whole aorta region
                entire_region = stats.get("entire_region")
                for measure in measure_keys:
                    if measure not in stats:
                        continue
                    measure_stats = stats[measure]
                    rows.append(
                        {
                            **identifier,
                            "region": region,
                            "measure": measure,
                            "mean_diameter_mm": measure_stats.get("mean_diameter"),
                            "major_diameter_mm": measure_stats.get("major_diameter"),
                            "minor_diameter_mm": measure_stats.get("minor_diameter"),
                            "area_mm2": measure_stats.get("area"),
                            "roundness": measure_stats.get("roundness"),
                            "flatness": measure_stats.get("flatness"),
                            "rel_distance": measure_stats.get("rel_distance"),
                            "entire_region": entire_region,
                        }
                    )
            return rows

        def _format_tissues_volumetric() -> list[dict]:
            volumetric = self.stats.get("tissues", {}).get("volumetric", {})
            if not volumetric:
                logger.warning(f"No tissue volumetric stats found in {self.path}")
                return []
            rows = []
            for region, stats in volumetric.items():
                for structure, structure_stats in stats.items():
                    rows.append(
                        {
                            **identifier,
                            "region": region,
                            "structure": structure,
                            "volume_cm3": structure_stats.get("volume_cm3"),
                            "hu_mean": structure_stats.get("hu_mean"),
                            "hu_std_dev": structure_stats.get("hu_std_dev"),
                        }
                    )
            return rows

        def _format_tissues_vertebral() -> list[dict]:
            vertebral = self.stats.get("tissues", {}).get("vertebral", {})
            if not vertebral:
                logger.warning(f"No tissue vertebral stats found in {self.path}")
                return []
            rows = []
            for vertebra, stats in vertebral.items():
                for structure, structure_stats in stats.items():
                    for measure, value in structure_stats.items():
                        rows.append(
                            {
                                **identifier,
                                "vertebra": vertebra,
                                "structure": structure,
                                "measurement": measure,
                                "value": value,
                            }
                        )
            return rows

        for table in stats_tables:
            match table:
                case "metadata":
                    table_data = _format_metadata()
                case "vol_int":
                    table_data = _format_vol_int()
                case "contrast":
                    table_data = _format_contrast()
                case "vertebrae":
                    table_data = _format_vertebrae()
                case "aorta_metrics":
                    table_data = _format_aorta_metrics()
                case "aorta_diameters":
                    table_data = _format_aorta_diameters()
                case "tissues_volumetric":
                    table_data = _format_tissues_volumetric()
                case "tissues_vertebral":
                    table_data = _format_tissues_vertebral()
            formatted_stats[table] = table_data
        self.db_stats = formatted_stats
        return self


def resample_with_sitk(
    image,
    new_spacing=None,
    new_size=None,
    is_label=False,
    interpolator_type="gaussian",
    reference_image=None,
    reorient=None,
) -> sitk.Image:
    """Resample a SimpleITK image to a new spacing, size, or reference image.
    :param image: The SimpleITK image to resample.
    :param new_spacing: The new spacing to resample to. If None, the original
        spacing will be used to calculate the new size.
    :param new_size: The new size to resample to. If None, the original
        size will be used to calculate the new spacing.
    :param is_label: Whether the image is a label image. If True, the
        resampling will use label-specific interpolators.
    :param interpolator_type: The type of interpolator to use for resampling.
        Options are 'gaussian', 'linear', 'nearest' for label images, and
        'lanczos', 'bspline', 'gaussian', 'linear' for regular images.
    :param reference_image: An optional reference image to use for resampling.
        If provided, the new spacing and size will be taken from this image.
    :param reorient: The CT orientation to reorient the image to. If None, no reorientation is applied.
    :return: The resampled SimpleITK image."""
    # Get original properties
    original_spacing = image.GetSpacing()
    original_size = image.GetSize()

    resampler = sitk.ResampleImageFilter()
    if reference_image is not None:
        resampler.SetReferenceImage(reference_image)
    else:
        if new_spacing is not None and new_size is not None:
            raise ValueError("Only one of new_spacing and new_size may be specified")
        # Calculate new size based on spacing change
        if new_spacing is not None:
            new_size = [
                int(original_size[i] * original_spacing[i] / new_spacing[i])
                for i in range(3)
            ]
        elif new_size is not None:
            new_spacing = [
                original_spacing[i] * original_size[i] / new_size[i] for i in range(3)
            ]
        else:
            raise ValueError("Must specify either a new_size or new_spacing")
        #    Set up resampling parameters
        resampler.SetOutputSpacing(new_spacing)
        resampler.SetSize(new_size)
        resampler.SetOutputDirection(image.GetDirection())
        resampler.SetOutputOrigin(image.GetOrigin())
        resampler.SetTransform(sitk.Transform())
        resampler.SetDefaultPixelValue(0)

    if is_label:
        if interpolator_type == "gaussian":
            interpolator = sitk.sitkLabelGaussian
        elif interpolator_type == "linear":
            interpolator = sitk.sitkLabelLinear
        elif interpolator_type == "nearest":
            interpolator = sitk.sitkNearestNeighbor
        else:
            raise ValueError(
                f"label interpolator must be in [gaussian, linear, nearest], got {interpolator_type}"
            )
        resampler.SetInterpolator(interpolator)
    else:
        if interpolator_type == "lanczos":
            interpolator = sitk.sitkLanczosWindowedSinc
        elif interpolator_type == "bspline":
            interpolator = sitk.sitkBSpline
        elif interpolator_type == "gaussian":
            interpolator = sitk.sitkGaussian
        elif interpolator_type == "linear":
            interpolator = sitk.sitkLinear
        else:
            raise ValueError(
                f"image interpolator must be in [lanczos, gaussian, linear], got {interpolator_type}"
            )
        resampler.SetInterpolator(interpolator)
    resampled_image = resampler.Execute(image)
    if reorient is not None:
        current_orientation = (
            sitk.DICOMOrientImageFilter_GetOrientationFromDirectionCosines(
                resampled_image.GetDirection()
            )
        )
        if current_orientation != reorient:
            logger.debug(f"Reorienting image from {current_orientation} to {reorient}.")
            reorienter = sitk.DICOMOrientImageFilter()
            reorienter.SetDesiredCoordinateOrientation(reorient)
            resampled_image = reorienter.Execute(resampled_image)
    return resampled_image


def load_nifti_paths(input_arg: Path, data_dir: Path) -> list[Path]:
    """Store either the contents of the input text file or the input nifti as a list of nifti paths
    Parameters
        input_arg: Path - the input argument path
        data_dir: Path - the data directory to prepend to the nifti file name {data_dir} / {nifti_path}
    Returns:
        A list of paths containing the nifti files in the input
    """
    single_file_suffixes = set([".nii", ".gz"])
    # Input arg must exist
    if not input_arg.exists():
        raise FileNotFoundError(
            f"Input argument {input_arg} does not exist. Please check path."
        )
    # If it's a single file - just prepend the data dir and return
    if set(input_arg.suffixes).intersection(single_file_suffixes):
        logger.info(f"Found input nifti {data_dir / input_arg}")
        return [data_dir / input_arg]
    # Otherwise we load the file and validate
    with input_arg.open() as f:
        niftis = [data_dir / Path(line.strip()) for line in f.readlines()]
    if len(niftis) == 0:
        raise ValueError(
            f"Nifti list file {input_arg} is empty. Please check contents."
        )
    logger.info(f"Found {len(niftis)} nifti files to analyze in {input_arg}")
    return niftis


class NotNiftiFileError(ValueError):
    pass


class SimpleITKReadError(ValueError):
    pass



from .converter import DicomConverter
from pathlib import Path


def convert_dicoms_to_nifti(convert_args):
    """Convert dicom files to NIfTI format.
    Args:
        convert_args (argparse.Namespace): Arguments from the command line.
    """
    converter = DicomConverter(
        axial_only=convert_args.axial_only,
        no_mip=convert_args.no_mip,
        resample=not convert_args.no_resampling,
        resample_spline_interpolation_order=convert_args.resample_interpolation_order,
        resample_padding=convert_args.resample_padding,
        validate_orthogonal=convert_args.validate_orthogonal,
        validate_slice_increment=not convert_args.skip_slice_increment_validation,
        validate_instance_number=not convert_args.skip_instance_number_validation,
        validate_slice_count=not convert_args.skip_slice_count_validation,
        n_processes=convert_args.n_processes,
        threads_per_process=convert_args.threads_per_process,
        min_slice_count=convert_args.min_slice_count,
        db_batch_size=convert_args.db_batch_size,
    )
    if convert_args.dicoms.is_file():
        # If the input is a file, read the list of dicom directories from the file
        with convert_args.dicoms.open() as f:
            dicom_folders = [
                convert_args.data_dir / Path(line.strip()) for line in f.readlines()
            ]
    elif convert_args.dicoms.is_dir():
        # If the input is a directory, use it as the list of dicom directories
        dicom_folders = [convert_args.data_dir / convert_args.dicoms]
    else:
        raise ValueError(
            f"Input dicoms argument should be either a text file containing lists of dicom folders, or a dicom folder itself"
        )
    converter.convert(
        dicom_folders=dicom_folders,
        output_directory=convert_args.output_dir,
    )


def add_dicom_conversion_subparser(subparsers):
    # Add subcommands
    convert_parser = subparsers.add_parser(
        "convert", help="Convert dicom data to the NIfTI format"
    )
    convert_parser.add_argument(
        "dicoms",
        help="Path to a dicom directory or a file containing a list of dicom directories",
        type=Path,
    )
    convert_parser.add_argument(
        "output_dir",
        type=Path,
        help="Path to the output directory",
    )
    convert_parser.add_argument(
        "--data-dir",
        "-d",
        type=Path,
        default=Path(""),
        help="Data directory for the input dicoms (i.e. {data_dir}/dicom_folder) - useful for remapping in docker. Default is no data directory.",
    )
    convert_parser.add_argument(
        "--n-processes",
        "-n",
        type=int,
        default=1,
        help="Number of processes to use for conversion (default: 1)",
    )
    convert_parser.add_argument(
        "--threads-per-process",
        "-t",
        type=int,
        default=4,
        help="Maximum number of transformation threads per process (default: 4)",
    )
    convert_parser.add_argument(
        "--axial-only",
        "-ax",
        action="store_true",
        help="Only convert axial series",
    )
    convert_parser.add_argument(
        "--no-mip", "-nm", action="store_true", help="Do not convert MIP series"
    )
    convert_parser.add_argument(
        "--no-resampling",
        "-nr",
        action="store_true",
        help="Do not resample the dicoms to LAS orientation",
    )
    convert_parser.add_argument(
        "--resample-interpolation-order",
        "-i",
        type=int,
        default=3,
        help="Interpolation order for resampling (default: 3)",
    )
    convert_parser.add_argument(
        "--resample-padding",
        "-p",
        type=int,
        default=-1024,
        help="Padding Hounsfield Unit for resampling (default: -1024)",
    )
    convert_parser.add_argument(
        "--validate-orthogonal",
        "-v",
        action="store_true",
        help="Validate that the dicoms are orthogonal. Will not convert gantry tilted series.",
    )
    convert_parser.add_argument(
        "--skip-slice-increment-validation",
        "-ssi",
        action="store_true",
        help="Skip validation of consistent slice increment in DICOM series. Use with caution",
    )
    convert_parser.add_argument(
        "--skip-instance-number-validation",
        "-sin",
        action="store_true",
        help="Skip validation of dicom instance numbers. Use with caution.",
    )
    convert_parser.add_argument(
        "--skip-slice-count-validation",
        "-ssc",
        action="store_true",
        help="Skip validation of slice counts in the dicoms. Use with caution.",
    )
    convert_parser.add_argument(
        "--min-slice-count",
        "-s",
        type=int,
        default=30,
        help="Minimum number of slices to consider a series valid (default: 30)",
    )
    convert_parser.add_argument(
        "--db-batch-size",
        "-b",
        type=int,
        default=100,
        help="Batch size for database insertion (default: 100)",
    )



import SimpleITK as sitk
import numpy as np
import xs3d

from kimimaro import skeletonize
from skimage.filters import gaussian
from skimage.morphology import remove_small_holes, remove_small_objects
from loguru import logger
from mircat_v2.stats.centerline import Centerline
from mircat_v2.stats.utilities import filter_segmentation, get_regionprops


class Vessel:
    """A class to represent a single vessel with its centerline and properties."""

    def __init__(
        self,
        segmentation,
        label,
        label_map,
        remap=True,
        largest_connected_component=True,
    ):
        """
        Initialize the Vessel class with segmentation data and label map.
        :param segmentation: The segmentation data for the vessel.
        :param label: The label of the vessel to be processed.
        :param label_map: A dictionary mapping structure names to label IDs.
        :param remap: Whether to remap the labels in the segmentation.
        :param largest_connected_component: Whether to filter to the largest connected component for the label.
        """
        self.segmentation = filter_segmentation(
            segmentation,
            label,
            label_map,
            remap=remap,
            largest_connected_component=largest_connected_component,
        )
        self.anisotropy = list(self.segmentation.GetSpacing())
        self.array = sitk.GetArrayFromImage(self.segmentation)
        self.label = label
        self.label_idx = 1 if remap else label_map.get(label, 0)
        if remap:
            self.label_map = {1: label}
        else:
            self.label_map = {label_map.get(label, 0): label}
        self.skeletonization_kwargs = {
            "teasar_params": {
                "scale": 1.0,
                "const": 40,
            },
            "object_ids": None,
            "anisotropy": self.anisotropy,
            "dust_threshold": 1000,
            "progress": False,
            "fix_branching": True,
            "in_place": False,
            "fix_borders": True,
            "parallel": 1,
            "parallel_chunk_size": 100,
            "extra_targets_before": [],
            "extra_targets_after": [],
            "fill_holes": False,
            "fix_avocados": False,
            "voxel_graph": None,
        }
        self.skeleton = None
        self.centerline = None
        self.cpr = None
        self.cpr_stats = None

    @property
    def length(self):
        """
        Calculate the length of the vessel's centerline in physical space.
        :return: The length of the centerline in physical units.
        """
        if not hasattr(self, "centerline"):
            raise ValueError(
                "Centerline has not been created. Call create_centerline() first."
            )
        # self.centerline.length is voxel space - need to convert with spacing
        return round(self.centerline.length * min(self.anisotropy), 1)

    def create_skeleton(self, **kwargs):
        """
        Create a centerline skeleton for the vessel.
        :param kwargs: Additional keyword arguments for kimimaro skeletonization.
        :return: The Vessel instance with the skeleton created.
        """
        self.skeletonization_kwargs.update(kwargs)
        skeletons = skeletonize(self.array, **self.skeletonization_kwargs)
        self.skeleton = skeletons[self.label_idx]
        return self

    def create_centerline(self, **kwargs):
        """
        Create a centerline for the vessel.
        :param kwargs: Additional keyword arguments for kimimaro skeletonization.
        :return: The Vessel instance with the centerline created.
        """
        if self.skeleton is None:
            logger.debug("Skeleton not created yet. Creating skeleton first.")
            self.create_skeleton(**kwargs)
            if self.skeleton is None:
                raise ValueError(
                    "Skeleton for the vessel is None. Ensure skeletonization was successful."
                )
        # This orders the centerline in physical space
        coordinates = self.skeleton.paths()[0]
        coordinates = (coordinates / self.anisotropy).round().astype(int)
        centerline = Centerline(
            label=self.label_idx, vessel=self.label, coordinates=coordinates
        )
        self.centerline = centerline
        return self

    def process_centerline(
        self, target_points: int, smoothing_factor: float = 1.0, degree: int = 3
    ):
        """
        Postprocess the centerline by resampling it with spline interpolation to a desired number of points and calculate tangent vectors.
        :param target_points: Number of points in the resampled centerline.
        :param smoothing_factor: Smoothing factor for spline interpolation.
        :param degree: Degree of the spline (1=linear, 2=quadratic, 3=cubic).
        """
        if not hasattr(self, "centerline"):
            raise ValueError(
                "Centerline has not been created. Call create_centerline() first."
            )
        self.centerline.resample_with_spline(
            target_points=target_points, smoothing=smoothing_factor, degree=degree
        )
        self.centerline.calculate_tangent_vectors()
        return self

    def create_straightened_cpr(self, radius: int = 100):
        """
        Create a straightened CPR (Curved Planar Reformation) from the vessel's centerline.
        :param radius: The radius of the sliced image in physical units of the scan (e.g., 50 = 50mm radius if spacing is in mm).
        :return: A 2D numpy array representing the straightened CPR.
        """
        if not hasattr(self, "centerline"):
            raise ValueError(
                "Centerline has not been created. Call create_centerline() first."
            )
        cpr = create_straightened_cpr_from_array(
            self.array,
            self.centerline.coordinates,
            self.centerline.tangents,
            self.anisotropy,
            radius=radius,
            is_label=True,
        )
        assert len(cpr) == len(self.centerline.coordinates), (
            "CPR length does not match centerline coordinates length."
        )
        self.cpr = cpr
        return self

    def calculate_straightened_cpr_cross_section_stats(self):
        """
        Calculate statistics for each cross section of the CPR.
        Each cross section is analyzed for its mean diameter, major diameter, minor diameter, area, flatness, and roundness.
        :return: The Vessel instance with the CPR statistics calculated.
        """
        if self.cpr is None:
            raise ValueError(
                "CPR has not been created. Call create_straightened_cpr() first."
            )
        # Implement statistics calculation logic here
        cross_section_spacing = self.anisotropy[
            1:
        ]  # we assume the first dimension is the slice dimension
        stats = {
            idx: self.calculate_cross_section_stats(
                cross_section, pixel_spacing=cross_section_spacing
            )
            for idx, cross_section in enumerate(self.cpr)
        }
        assert len(stats) == len(self.cpr) == len(self.centerline), (
            "Stats length mismatch with CPR or centerline."
        )
        self.cpr_stats = stats
        return self

    @staticmethod
    def calculate_cross_section_stats(
        cross_section: np.ndarray, pixel_spacing: tuple | np.ndarray
    ) -> dict:
        """
        Calculate statistics for a single cross section.

        The "roundness" is calculated as (4 * pi * area) / (perimeter ** 2), which is a standard measure of how close the shape is to a perfect circle.

        """
        # This uses a custom get_regionprops wrapper (which internally uses skimage.measure.regionprops) to calculate the properties of the cross section
        regionprops = get_regionprops(cross_section)
        # This uses skimage.measure.regionprops to calculate the properties of the cross section
        regionprops = get_regionprops(cross_section)
        data_dict = {
            "mean_diameter": None,
            "major_diameter": None,
            "minor_diameter": None,
            "area": None,
            "flatness": None,
            "roundness": None,
            "centroid": None,
        }
        if len(regionprops) == 0:
            return data_dict
        region = regionprops[0]
        major_endpoints, minor_endpoints = _get_cross_section_diameter_endpoints(region)
        major_endpoints = [np.array(pt) * pixel_spacing for pt in major_endpoints]
        minor_endpoints = [np.array(pt) * pixel_spacing for pt in minor_endpoints]
        major_diameter = np.linalg.norm(
            np.array(major_endpoints[1]) - np.array(major_endpoints[0])
        ).round(1)
        minor_diameter = np.linalg.norm(
            np.array(minor_endpoints[1]) - np.array(minor_endpoints[0])
        ).round(1)
        data_dict = {
            "mean_diameter": np.mean([major_diameter, minor_diameter]).round(1),
            "major_diameter": major_diameter,
            "minor_diameter": minor_diameter,
            "area": region.area * np.prod(pixel_spacing),
            "flatness": round(major_diameter / minor_diameter, 3)
            if minor_diameter > 0
            else 0,
            "roundness": round((4 * np.pi * region.area) / (region.perimeter**2), 3)
            if region.perimeter > 0
            else 0,
            "centroid": tuple(round(x) for x in region.centroid),
        }
        return data_dict

    def _plot_centerline(self):
        """
        Plot the centerline of the vessel against segmentation in two axes. Meant for debugging.
        """
        if self.centerline is None:
            raise ValueError(
                "Centerline has not been created. Call create_centerline() first."
            )
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 2, figsize=(12, 6))
        ax[0].imshow(self.array.max(axis=2), cmap="gray")
        ax[0].plot(
            self.centerline.coordinates[:, 1],
            self.centerline.coordinates[:, 0],
            color="red",
            linewidth=2,
        )
        ax[0].set_title(f"Saggital View of {self.label}")
        ax[1].imshow(self.array.max(axis=1), cmap="gray")
        ax[1].plot(
            self.centerline.coordinates[:, 2],
            self.centerline.coordinates[:, 0],
            color="red",
            linewidth=2,
        )
        ax[1].set_title(f"Coronal View of {self.label}")
        plt.tight_layout()
        plt.show()


def _get_cross_section_diameter_endpoints(region) -> tuple[tuple, tuple]:
    """
    Get the major and minor diameter endpoints of a cross section from a regionprops object.
    :param region: A regionprops object containing the properties of the region.
    :return: A tuple containing the major and minor diameter endpoints.
    The endpoints are tuples of (y, x) coordinates.
    """

    def _get_axis_endpoints(centroid, orientation, axis_length):
        y0, x0 = centroid
        # calculate the endpoints of the major axis using the centroid
        x1 = x0 - np.sin(orientation) * 0.5 * axis_length
        x2 = x0 + np.sin(orientation) * 0.5 * axis_length
        y1 = y0 - np.cos(orientation) * 0.5 * axis_length
        y2 = y0 + np.cos(orientation) * 0.5 * axis_length
        return (y1, x1), (y2, x2)

    centroid = region.centroid
    orientation = region.orientation
    major_axis_length = region.major_axis_length
    minor_axis_length = region.minor_axis_length
    major_endpoints = _get_axis_endpoints(centroid, orientation, major_axis_length)
    minor_endpoints = _get_axis_endpoints(centroid, orientation, minor_axis_length)
    return major_endpoints, minor_endpoints


def create_straightened_cpr_from_array(
    array: np.ndarray,
    centerline: np.ndarray,
    tangents: np.ndarray,
    anisotropy: tuple | list,
    radius: int = 200,
    is_label: bool = True,
) -> np.ndarray:
    """
    Create a straightened CPR (Curved Planar Reformation) of an array from a centerline and relevant tangent vectors.
    :param array: The 3D numpy array to slice.
    :param centerline: numpy array containing the coordinates of the centerline.
    :param tangents: numpy array containing the tangent vectors at each point of the centerline
    :param anisotropy: A tuple or list representing the spacing of the image (e.g., (1.0, 1.0, 1.0)).
    :param radius: The radius of the sliced image in physical units of the scan (i.e. 50 = 50mm radius if spacing is in mm). 200 units is default.
    :return: A 2D numpy array representing the straightened CPR.
    """
    array = np.asfortranarray(array)
    expected_length = int(np.ceil(radius / min(anisotropy))) * 2
    expected_size = (expected_length, expected_length)
    pad_value = np.min(array)
    min_size = int(
        round(25 / min(anisotropy))
    )  # Minimum for a segmentation allowed in the cross section
    cpr = []
    for point, tangent in zip(centerline, tangents):
        cross_section = xs3d.slice(
            array, point, tangent, anisotropy, standardize_basis=True, crop=radius
        )
        cross_section = _resize_cross_section(cross_section, expected_size, pad_value)
        if is_label:
            cross_section = _postprocess_label_cross_section(cross_section, min_size)
        cpr.append(cross_section)
    cpr = np.array(cpr, dtype=array.dtype)
    return cpr


def _resize_cross_section(
    cross_section: np.ndarray, target_size: tuple, pad_value: float | int
) -> np.ndarray:
    """
    Resize cross_section to target_size by padding or cropping from center.

    :param cross_section: 2D array to resize
    :param target_size: Target (height, width)
    :return: Resized array
    """
    current_h, current_w = cross_section.shape
    target_h, target_w = target_size

    # Calculate padding/cropping for each dimension
    h_diff = target_h - current_h
    w_diff = target_w - current_w

    # Start with the original cross section
    result = cross_section

    # Handle height dimension
    if h_diff > 0:
        # Need to pad
        pad_top = h_diff // 2
        pad_bottom = h_diff - pad_top
        result = np.pad(
            result,
            ((pad_top, pad_bottom), (0, 0)),
            mode="constant",
            constant_values=pad_value,
        )
    elif h_diff < 0:
        # Need to crop
        crop_top = (-h_diff) // 2
        crop_bottom = crop_top + target_h
        result = result[crop_top:crop_bottom, :]

    # Handle width dimension
    if w_diff > 0:
        # Need to pad
        pad_left = w_diff // 2
        pad_right = w_diff - pad_left
        result = np.pad(
            result,
            ((0, 0), (pad_left, pad_right)),
            mode="constant",
            constant_values=pad_value,
        )
    elif w_diff < 0:
        # Need to crop
        crop_left = (-w_diff) // 2
        crop_right = crop_left + target_w
        result = result[:, crop_left:crop_right]

    return result


def _postprocess_label_cross_section(
    cross_section: np.ndarray, min_size: int
) -> np.ndarray:
    """
    Postprocess a label cross-section by removing small holes and smoothing.

    :param cross_section: 2D array representing the label cross-section
    :return: Processed 2D array
    """
    labels = [label for label in np.unique(cross_section) if label > 0]
    if len(labels) == 0:
        return cross_section
    output_cross_section = np.zeros_like(cross_section, dtype=np.uint8)
    for label in labels:
        mask = (cross_section == label).astype(np.uint8)
        regionprops = get_regionprops(mask)
        if len(regionprops) == 0:
            centered_label = np.zeros_like(mask)
        elif len(regionprops) > 1:
            centered_label = _closest_to_centroid(mask, regionprops)
        else:
            centered_label = mask
        centered_label = centered_label.astype(bool)
        centered_label = remove_small_holes(centered_label)
        centered_label = remove_small_objects(centered_label, min_size=10)
        centered_label = gaussian(centered_label, sigma=2).round().astype(np.uint8)
        output_cross_section[centered_label == 1] = label
    return output_cross_section


def _closest_to_centroid(cross_section: np.ndarray, regions: list) -> np.ndarray:
    """Filter a cross-section label using skimage.measure.regionprops to the region closest
    to the center
    :param cross_section: the cross-section array
    :param regions: the list output from skimage.measure.regionprops
    :return: the filtered numpy array
    """
    center_of_plane = np.array(cross_section.shape) / 2.0
    centroids = [np.array(region.centroid) for region in regions]
    distance_per_region = np.asarray(
        [np.linalg.norm(centroid - center_of_plane) for centroid in centroids]
    )
    min_distance_region_idx = int(np.argmin(distance_per_region))
    center_region = regions[min_distance_region_idx]
    center_label = np.zeros_like(cross_section, dtype=np.uint8)
    center_label[center_region.coords[:, 0], center_region.coords[:, 1]] = 1
    return center_label



import numpy as np
import SimpleITK as sitk
from skimage.measure import label, regionprops


def calculate_shape_stats(seg: sitk.Image) -> sitk.LabelShapeStatisticsImageFilter:
    """Calculate the shape stats for a segmentation
    Parameters
    ----------
    seg : sitk.Image
        The segmentation to calculate shape stats for
    Returns
    -------
    sitk.LabelShapeStatisticsImageFilter
        The executed shape stats
    """
    shape_stats = sitk.LabelShapeStatisticsImageFilter()
    shape_stats.SetGlobalDefaultCoordinateTolerance(1e-5)
    shape_stats.ComputeOrientedBoundingBoxOn()
    shape_stats.ComputePerimeterOn()
    shape_stats.Execute(seg)
    return shape_stats


def calculate_intensity_stats(
    image: sitk.Image, seg: sitk.Image
) -> sitk.LabelIntensityStatisticsImageFilter:
    """Calculate intensity stats for a segmentation using the reference image
    Parameters
    ----------
    image : sitk.Image
        The reference image
    seg : sitk.Image
        The segmentation of the reference image
    Returns
    -------
    sitk.LabelIntensityStatisticsImageFilter
        The executed intensity stats
    """
    intensity_stats = sitk.LabelIntensityStatisticsImageFilter()
    intensity_stats.SetGlobalDefaultCoordinateTolerance(1e-5)
    intensity_stats.Execute(seg, image)
    return intensity_stats


def filter_segmentation(
    segmentation: sitk.Image,
    labels: str | list[str],
    label_map: dict[str, int],
    largest_connected_component: bool = False,
    remap: bool = False,
) -> sitk.Image:
    """Filter a segmentation to only include specified labels
    Parameters
    ----------
    segmentation : sitk.Image
        The segmentation to filter
    labels : str|list[str]
        The labels to keep in the segmentation
    label_map : dict[str, int]
        A mapping of structure names to label IDs
    remap : bool, optional
        If True, remap the labels indexed starting from 1, by default False
    Returns
    -------
    sitk.Image
        The filtered segmentation
    """
    filtered_segmentation = sitk.Image(segmentation.GetSize(), sitk.sitkUInt8)
    filtered_segmentation.CopyInformation(segmentation)
    if isinstance(labels, str):
        labels = [labels]
    for i, label in enumerate(labels, start=1):
        idx = label_map.get(label, None)
        if idx is not None:
            mask = sitk.BinaryThreshold(segmentation, idx, idx, 1, 0)
            if largest_connected_component:
                mask = get_largest_connected_component(mask)
            new_idx = i if remap else idx
            mask = mask * new_idx
            filtered_segmentation = sitk.Add(filtered_segmentation, mask)
    return filtered_segmentation


def get_largest_connected_component(binary_segmentation: sitk.Image) -> sitk.Image:
    """Get the largest connected component from a binary segmentation
    Parameters
    ----------
    binary_segmentation : sitk.Image
        The binary segmentation to get the largest connected component from
    Returns
    -------
    sitk.Image
        The largest connected component of the binary segmentation
    """
    cc_filter = sitk.ConnectedComponentImageFilter()
    labeled = cc_filter.Execute(binary_segmentation)
    relabel_filter = sitk.RelabelComponentImageFilter()
    relabeled = relabel_filter.Execute(labeled)
    largest_component = sitk.BinaryThreshold(relabeled, 1, 1, 1, 0)
    return largest_component


def get_regionprops(image) -> list:
    """Get region properties from a labeled image
    Parameters
    ----------
    image : np.ndarray
        The labeled image to get region properties from
    Returns
    -------
    list
        A list of region properties for each labeled region in the image
    """
    labeled_image = label(image)
    return regionprops(labeled_image)


def slice_images(
    endpoints: tuple, images: sitk.Image | list
) -> sitk.Image | list[sitk.Image]:
    """Slice any number of matching images to the same region using endpoints
    Parameters:
        endpoints: Specific z-axis endpoints for a region of interest in the shape (start, end). Default is (0, None),
                    which will measure the entire image. If end = None, will measure from start_idx -> rest of the scan
        images: a single sitk.Image or list of sitk.Images -> must be the same shape or error is raised
    Returns:
        A single sitk.Image or list of sitk.Images sliced to the specified endpoints.
    """
    start, end = endpoints
    is_single_image = isinstance(images, sitk.Image)
    if end is None:
        if is_single_image:
            end = images.GetSize()[
                -1
            ]  # This is the full length of the image along the z-axis
        else:
            # Check that all images are the same size
            image_sizes = [img.GetSize() for img in images]
            assert image_sizes.count(image_sizes[0]) == len(image_sizes), ValueError(
                "All images in list must be the same shape"
            )
            end = images[0].GetSize()[-1]  # All images are assumed to be the same size
    if is_single_image:
        images = images[:, :, start:end]
        return images
    else:
        sliced_images = [image[:, :, start:end] for image in images]
        return sliced_images


def calculate_3d_volumes(
    segmentation: sitk.Image,
    label_map: dict[str, int],
    endpoints: tuple = (0, None),
    unit: str = "mm3",
) -> dict[str, float]:
    """Calculate the 3D volumes of each label in a segmentation
    Parameters:
        segmentation: The segmentation to calculate volumes for
        label_map: A mapping of structure names to label IDs
        endpoints: Specific z-axis endpoints for a region of interest in the shape (start, end). Default is (0, None),
                    which will measure the entire image. If end = None, will measure from start_idx -> rest of the scan
        dividend: A value to divide the calculated volumes by, default is 1.0 (no change) - can be used to convert units
    Returns:
        A dictionary mapping structure names to their 3D volumes the native units of the segmentation
    """
    segmentation = slice_images(endpoints, segmentation)
    shape_stats = calculate_shape_stats(segmentation)
    seg_labels = shape_stats.GetLabels()
    volumes = {}
    if unit == "cm3":
        dividend = 1000.0  # Convert from mm^3 to cm^3
        decimal_places = 1
    elif unit == "L":
        dividend = 1000000.0  # Convert from mm^3 to L
        decimal_places = 3
    else:
        dividend = 1.0
        decimal_places = 1
    for name, label in label_map.items():
        if label in seg_labels:
            volume = shape_stats.GetPhysicalSize(label)
            volumes[f"{name}"] = {
                f"volume_{unit}": round(volume / dividend, decimal_places)
            }
    return volumes


def calculate_3d_intensities(
    image: sitk.Image,
    segmentation: sitk.Image,
    label_map: dict[str, int],
    endpoints: tuple = (0, None),
) -> dict[str, float]:
    """Calculate the average intensities of each label in a segmentation
    Parameters:
        image: The reference image to calculate intensities from
        segmentation: The segmentation to calculate volumes for
        label_map: A mapping of structure names to label IDs
        endpoints: Specific z-axis endpoints for a region of interest in the shape (start, end). Default is (0, None),
                    which will measure the entire image. If end = None, will measure from start_idx -> rest of the scan
    Returns:
        A dictionary mapping structure names to their 3D intensities in Hounsfield Units
    """
    image = slice_images(endpoints, image)
    segmentation = slice_images(endpoints, segmentation)
    instensity_stats = calculate_intensity_stats(image, segmentation)
    seg_labels = instensity_stats.GetLabels()
    intensities = {}
    for name, label in label_map.items():
        if label in seg_labels:
            mean_hu = instensity_stats.GetMean(label)
            std_dev_hu = instensity_stats.GetStandardDeviation(label)
            intensities[f"{name}"] = {
                "hu_mean": round(mean_hu, 1),
                "hu_std_dev": round(std_dev_hu, 1),
            }
    return intensities


def calculate_2d_areas(
    seg: sitk.Image,
    label_map: dict,
    slice_idx: int,
    get_perimeter: bool = False,
    units: str = "cm",
) -> dict:
    """Calculate the area of each label in the segmentation within a 2d slice
    :param seg: the segmentation
    :param label_map: the label map
    :param slice_idx: the slice of the segmentation to measure
    :param prefix: the label of the slice (L1, L3, etc)
    :param get_perimeter: Measure the perimeters from the shape stats
    :param units: the base unit to use for the area and perimeter, default is 'cm'
    :return: a dictionary containing the area of each label found within the segmentation
    """
    seg_slice = seg[:, :, slice_idx]
    shape_stats = calculate_shape_stats(seg_slice)
    seg_labels = shape_stats.GetLabels()
    if units == "cm":
        area_div = 100.0
        linear_div = 10.0
    elif units == "mm":
        area_div = 1.0
        linear_div = 1.0
    data = {}
    for name, label in label_map.items():
        # Don't want the background label
        if label == 0 or label not in seg_labels:
            continue
        data[name] = {}
        area = round(shape_stats.GetPhysicalSize(label) / area_div, 1)
        data[name][f"area_{units}2"] = area
        if get_perimeter and (label in seg_labels):
            # Note that the shape touches the border if at least 5 percent of the perimeter is on the border
            border_ratio = shape_stats.GetPerimeterOnBorderRatio(label) * 100
            raw_perim = (
                shape_stats.GetPerimeter(label) / linear_div
            )  # This will be perimeter in cm
            ellipse_perim = _calc_ellipsoid_perimeter(
                shape_stats.GetEquivalentEllipsoidDiameter(label), units
            )
            circ_perim = shape_stats.GetEquivalentSphericalPerimeter(label) / linear_div
            # Use colon to separate the name and the value
            perimeter_data = {
                "border_ratio": round(border_ratio, 1),
                f"raw_perimeter_{units}": round(raw_perim, 1),
                f"ellipse_perimeter_{units}": round(ellipse_perim, 1),
                f"circle_perimeter_{units}": round(circ_perim, 1),
            }
            data[name].update(perimeter_data)
    return data


def _calc_ellipsoid_perimeter(diameters: tuple, unit="cm") -> float:
    """Calculate the estimated ellipsoid diameters from a shape_stats.GetEquivalentEllipsoidDiameter result.
    Uses the Ramanujan estimation.
    :param diameters: the result from shape_stats
    :param unit: the unit of the diameters, default is 'cm'
    :return the perimeter in the specified unit
    """
    minor, major = diameters
    major /= 2
    minor /= 2
    perim = np.pi * (
        3 * (major + minor) - np.sqrt((3 * major + minor) * (major + 3 * minor))
    )
    if unit == "cm":
        dividend = 10.0
    elif unit == "mm":
        dividend = 1.0
    return perim / dividend



from dataclasses import dataclass
from pathlib import Path
import csv
import gzip
import numpy as np
import SimpleITK as sitk

from loguru import logger
from skimage import draw
from mircat_v2.stats.utilities import filter_segmentation, get_regionprops
from mircat_v2.stats.vessel import Vessel, create_straightened_cpr_from_array


def calculate_aorta_stats(nifti, seg_id: str, label_map: dict) -> dict:
    """Calculate aorta specific statistics for a given nifti file.
    :param nifti: StatsNifti object containing the nifti file and its metadata.
    :param seg_id: The segmentation ID to calculate statistics for.
    :param label_map: Dictionary mapping structure names to label IDs.
    :return: Dictionary containing aorta specific statistics.
    """
    logger.debug(f"Calculating aorta stats for {nifti.name}.")
    if not hasattr(nifti, "vertebrae_midlines"):
        logger.warning(
            f"Vertebrae midlines not found in {nifti.name}. Skipping aorta stats calculation."
        )
        return {}
    segmentation = nifti.id_to_seg_map.get(seg_id)
    aorta = Aorta(
        nifti.img,
        segmentation,
        nifti.vertebrae_midlines,
        label_map,
    )
    aorta.calculate_stats()
    aorta.write_aorta_summary(nifti.seg_folder / f"{nifti.name}_aorta_summary.csv.gz")
    return aorta.stats


@dataclass
class AortaStraightenedCPRCrossSection:
    """
    A dataclass to represent a cross-section of a straightened CPR.
    """

    index: int
    coordinates: np.ndarray
    thickness: float | None
    mean_diameter: float | None
    major_diameter: float | None
    minor_diameter: float | None
    area: float | None
    flatness: float | None
    roundness: float | None
    centroid: tuple[float, float] | None
    peria_total_volume: float
    peria_ring_volume: float
    peria_fat_volume: float
    peria_fat_hus: list[int]
    calc_volume_score: float
    calc_agatston_score: float
    calc_regions: int
    anatomical_region: str | None = None


class Aorta(Vessel):
    """Class for aorta-specific statistics and operations for the custom segmentation model specifically."""

    anatomical_regions_vertebrae_map = {
        "thoracic": [f"T{i}" for i in range(4, 13)],
        "desc": [f"T{i}" for i in range(5, 13)],
        "up_abd": ["T12", "L1", "L2"],
        "lw_abd": ["L2", "L3", "L4", "L5", "S1"],
    }
    # Assume that the aortic root is approximately 20mm in length
    root_length_mm: int = 20

    def __init__(
        self,
        image: sitk.Image,
        segmentation: sitk.Image,
        vertebrae_midlines: dict[str, int],
        label_map: dict,
    ):
        """
        Initialize the Aorta class with a nifti file and segmentation ID.

        Parameters:
            image: the ct scan in hounsfield units as a SimpleITK Image.
            segmentation: the segmentation containing the aorta as a SimpleITK Image.
            vertebrae_midlines: Dictionary mapping vertebrae names to their midline coordinates.
            label_map: Dictionary mapping structure names to label IDs.
        """
        # We need to include the brachiocephalic trunk and left subclavian artery to separate anatomical regions
        labels = ["aorta", "brachiocephalic_trunk", "subclavian_artery_left"]
        self.image = image
        self.segmentation = filter_segmentation(
            segmentation,
            labels=labels,
            label_map=label_map,
            remap=True,
            largest_connected_component=True,
        )
        self.anisotropy = list(self.segmentation.GetSpacing())
        self.array = sitk.GetArrayFromImage(self.segmentation)
        # We clip the hounsfield units to a range that is easier to see and useful for us
        self.image_array = sitk.GetArrayFromImage(self.image).clip(-400, 600)
        self.label_map = {label: idx for idx, label in enumerate(labels, start=1)}
        self.label = "aorta"
        self.label_idx = 1
        self.skeletonization_kwargs = {
            "teasar_params": {
                "scale": 1.0,
                "const": 40,
            },
            "object_ids": [1],  # Aorta is always label 1 in the segmentation
            "anisotropy": self.anisotropy,
            "dust_threshold": 1000,
            "progress": False,
            "fix_branching": True,
            "in_place": False,
            "fix_borders": True,
            "parallel": 1,
            "parallel_chunk_size": 100,
            "extra_targets_before": [],
            "extra_targets_after": [],
            "fill_holes": False,
            "fix_avocados": False,
            "voxel_graph": None,
        }
        self.vertebrae_midlines: dict = vertebrae_midlines
        # We will store these when needed - it does not need to be initialized immediately
        self.img_array: np.ndarray = np.array([])
        self.skeleton = None
        self.centerline = None
        self.cpr = np.array([])
        self.img_cprs = {}
        self.cross_section_stats: list[AortaStraightenedCPRCrossSection] = []
        self._regions: dict = None
        self.regions_by_index: dict = {}
        self.stats: dict = {}

    @property
    def regions(self):
        """
        Get the anatomical regions of the aorta.
        :return: Dictionary containing the anatomical regions and their properties.
        """
        if self._regions is None:
            self.determine_anatomical_regions()
        return self._regions

    def determine_anatomical_regions(self):
        """
        Determine the anatomical regions of the aorta in the segmentation based on vertebrae midlines.
        """
        regions = {}
        for region, vertebrae in self.anatomical_regions_vertebrae_map.items():
            midlines = [
                self.vertebrae_midlines[vert]
                for vert in vertebrae
                if self.vertebrae_midlines.get(vert) is not None
            ]
            # Need at least two midlines to determine a region
            if len(midlines) > 1:
                regions[region] = {
                    "in_image": True,
                    "entire_region": len(midlines) == len(vertebrae),
                    "start": min(midlines),
                    "end": max(midlines),
                }
            else:
                regions[region] = {
                    "in_image": False,
                    "entire_region": False,
                    "start": None,
                    "end": None,
                }
        # In order to properly measure the thoracic aorta - the entirety must be in the scan
        if regions["thoracic"]["in_image"] and regions["thoracic"]["entire_region"]:
            del regions["desc"]
        else:
            del regions["thoracic"]
        self._regions = regions
        return self

    def create_straightened_cprs(self, radius: int = 100):
        """
        Create both straightened CPRs of the segmentation and image from the centerline of the aorta.

        Parameters:
            radius: int - desired radius of the cross section in mm.
        Returns:
            self: The Aorta instance with the CPRs created.
        """
        # First calculate the segmentation cpr
        super().create_straightened_cpr(radius=radius)
        # Then calculate the image cpr
        img_cpr = create_straightened_cpr_from_array(
            self.image_array,
            self.centerline.coordinates,
            self.centerline.tangents,
            self.anisotropy,
            radius=radius,
            is_label=False,
        )
        # Assert that the image CPR matches the segmentation CPR and centerline
        assert len(img_cpr) == len(self.centerline.coordinates) == len(self.cpr), (
            "Image CPR length does not match centerline coordinates length."
        )
        # Boolean mask for fat voxels
        # periaortic fat should be between -190 and -30 HU
        fat = (-190 <= img_cpr) & (img_cpr <= -30)
        # Boolean mask for calcification
        # aortic calcification should be above 130 HU
        calcificiation = img_cpr >= 130
        self.img_cprs = {
            "image": img_cpr,
            "fat": fat,
            "calc": calcificiation,
        }
        return self

    # Overwriting the parent class method so we can use the AortaStraightenedCPRCrossSection dataclass
    def calculate_straightened_cpr_cross_section_stats(self):
        """
        Calculate statistics for each cross section of the CPR for the entirety of the aorta segmentation.
        Each cross section is analyzed for its mean diameter, major diameter, minor diameter, area, flatness, and roundness.
        :return: The Vessel instance with the CPR statistics calculated.
        """
        if self.cpr is None:
            raise ValueError(
                "CPR has not been created. Call create_straightened_cpr() first."
            )
        # we assume the first dimension is the slice dimension
        cross_section_spacing = self.anisotropy[1:]
        voxel_area = np.prod(cross_section_spacing)
        for idx, cross_section in enumerate(self.cpr):
            cross_section_stats = self.calculate_cross_section_stats(
                cross_section, cross_section_spacing
            )
            if idx == len(self.centerline.segment_lengths):
                # Estimate the final cross section thickness as the mean segment length
                thickness = self.centerline.segment_lengths.mean().round(2)
            else:
                thickness = self.centerline.segment_lengths[idx].round(2)
            periaortic_stats = self._calculate_cross_section_periaortic_fat(
                idx,
                voxel_area,
                thickness,
                cross_section_stats["mean_diameter"],
                cross_section_stats["centroid"],
            )
            calcification_stats = self._calculate_cross_section_calcification(
                idx, voxel_area, thickness
            )
            cross_section_stats.update(periaortic_stats)
            cross_section_stats.update(calcification_stats)
            cross_section_stats["index"] = idx
            cross_section_stats["coordinates"] = self.centerline.coordinates[idx]
            # cross_section_stats["array"] = cross_section
            cross_section_stats["thickness"] = thickness
            self.cross_section_stats.append(
                AortaStraightenedCPRCrossSection(**cross_section_stats)
            )
        assert len(self.cross_section_stats) == len(self.cpr) == len(self.centerline), (
            "Stats length mismatch with CPR or centerline."
        )
        return self

    def _calculate_cross_section_periaortic_fat(
        self,
        idx: int,
        area_per_voxel: float,
        thickness: float,
        diameter: float,
        centroid: tuple[int, int],
    ) -> dict:
        """
        Calculate the volume and collect the HUs of the periaortic fat in a cross section
        Parameters:
            idx: the index of the image and fat cpr to use
            thickness: the thickness of the cross section in mm
            diameter: the mean diameter of the segmentation cross section in mm
            centroid: the centroid of the segmentation cross section
        """
        if diameter is None or centroid is None:
            return {
                "peria_total_volume": 0,
                "peria_ring_volume": 0,
                "peria_fat_volume": 0,
                "peria_fat_hus": [],
            }
        # Get the appropriate slices
        seg_cs = self.cpr[idx] == self.label_map["aorta"]
        img_cs = self.img_cprs["image"][idx]
        fat_cs = self.img_cprs["fat"][idx]
        voxel_volume = area_per_voxel * thickness
        # Draw the disk around the centroid
        radius = (
            int(diameter / 2) + 10
        )  # we add 1cm to the radius when drawing the circle
        center_y, center_x = centroid
        # This is a boolean mask of the circle around the centroid
        ring_mask = np.zeros_like(img_cs, dtype=bool)
        # Draw the disk in the mask
        rr, cc = draw.disk((center_y, center_x), radius, shape=img_cs.shape)
        ring_mask[rr, cc] = True
        # Get the total volume of the mask before removing the aorta
        total_volume = np.sum(ring_mask) * voxel_volume
        # Remove the aorta from the mask
        ring_mask[seg_cs == 1] = False
        # Calculate the volume of the periaortic fat
        ring_volume = np.sum(ring_mask) * voxel_volume
        fat_volume = np.sum(fat_cs[ring_mask]) * voxel_volume
        # We get the fat_hu values as an array so that we can find the average for each aortic anatomical region
        fat_hus = img_cs[fat_cs & ring_mask]
        return {
            "peria_total_volume": total_volume,
            "peria_ring_volume": ring_volume,
            "peria_fat_volume": fat_volume,
            "peria_fat_hus": fat_hus,
        }

    def _calculate_cross_section_calcification(
        self, idx: int, voxel_area: float, thickness: float
    ) -> dict:
        """Calculate the calcification found within the cross section of the aorta
        Parameters:
            idx: the index of the cprs to measure
            voxel_area: the area of a voxel in the CPR in physical space (1mm, 1mm), (0.75mm, 0.75mm), etc.
            thickness: the thickness of the slice in mm
        Returns:
            a dictionary containing the derived calcification statistics
        """
        seg_cs = self.cpr[idx] == self.label_map["aorta"]
        if seg_cs.sum() == 0:
            return {"calc_volume_score": 0, "calc_agatston_score": 0, "calc_regions": 0}
        img_cs = self.img_cprs["image"][idx]
        calc_cs = self.img_cprs["calc"][idx]
        # Calcification has to be both >=130HU and within the aorta
        calc_in_aorta = calc_cs & seg_cs
        # Measure the calcification
        voxel_volume = voxel_area * thickness
        calc_volume_score = 0
        calc_agatston_score = 0
        calc_regions = get_regionprops(calc_in_aorta)
        for region in calc_regions:
            if region.area >= 1:
                # Volume score is just the amount of calcification in the cs
                calc_volume_score += region.area * voxel_volume
                # Agatston weights the area based on maximum HU inside the calcification
                max_calc_hu = img_cs[region.coords].max()
                assert max_calc_hu >= 130
                if 130 <= max_calc_hu < 200:
                    density_weighting_factor = 1
                elif 200 <= max_calc_hu < 300:
                    density_weighting_factor = 2
                elif 300 <= max_calc_hu < 400:
                    density_weighting_factor = 3
                else:
                    density_weighting_factor = 4
                calc_agatston_score += (
                    region.area * density_weighting_factor * voxel_area
                )
        return {
            "calc_volume_score": calc_volume_score,
            "calc_agatston_score": calc_agatston_score,
            "calc_regions": len(calc_regions),
        }

    def split_anatomical_regions_by_indicies(self) -> dict[str, list[int]]:
        """
        Split the aorta into [root, asc, arch, desc, up_abd, lw_abd] anatomical regions with reference to the indices of it's centerline/cpr.
        :return: Dictionary with region names as keys and lists of indices as values.
        """

        def _split_thoracic_aorta_by_index() -> dict[str, list[int]]:
            """Internal function to break apart the thoracic aorta"""
            # Check if both brachiocephalic trunk and left subclavian artery are in the cpr
            brach_label = self.label_map["brachiocephalic_trunk"]
            subclavian_label = self.label_map["subclavian_artery_left"]
            arch_start: int = None
            arch_end: int = None
            start, end = (
                self.regions["thoracic"]["start"],
                self.regions["thoracic"]["end"],
            )
            thoracic_indices = (
                self.centerline.get_indices_within_dimensional_coordinates(start, end)
            )
            # We use the anatomical method if the 2 extra segmentations were found
            if np.all(np.isin([brach_label, subclavian_label], self.cpr)):
                logger.debug("Using aortic CPR to define aortic arch")
                # We shrink this so we don't include the abdominal aorta
                thoracic_cpr = self.cpr[thoracic_indices]
                # The brachiocephalic trunk signifies the start of the aortic arch
                for idx, cross_section in enumerate(thoracic_cpr):
                    if brach_label in cross_section:
                        arch_start = idx
                        break
                # The left subclavian artery signifies the end of the aortic arch, so we go in reverse
                for idx, cross_section in enumerate(thoracic_cpr[::-1]):
                    if subclavian_label in cross_section:
                        arch_end = len(thoracic_cpr) - idx
            # Otherwise, we use the morphology of the aorta to define the regions
            else:
                logger.debug("Using aortic morphology to define aortic arch")
                # Need at least 50mm2 to be considered
                min_segmentation_area = 50
                # This is the peak of the centerline
                split = None
                arch_peak = int(self.centerline.coordinates[:, 0].min())
                aorta_array = self.array == self.label_map["aorta"]
                for idx, axial_slice in enumerate(self.array):
                    regions = get_regionprops(axial_slice)
                    # If there are two regions in the image, then we know the ascending and descending aorta
                    # are visible within the segmentation slice
                    if len(regions) == 2 and idx > arch_peak:
                        reg0 = regions[0]
                        reg1 = regions[1]
                        # If both sections of the aorta are sufficiently large,
                        if (
                            reg0.area > min_segmentation_area
                            and reg1.area > min_segmentation_area
                        ):
                            split = idx
                            break
                if split is None:
                    logger.error("Aortic arch could not be defined from morphology")
                    raise ValueError("Aortic arch could not be defined from morphology")
                # We shrink this to exclude the abdominal aorta
                thoracic_centerline = self.centerline.coordinates[thoracic_indices]
                for idx, coordinate in enumerate(thoracic_centerline):
                    if coordinate[0] <= split:
                        arch_start = idx
                        break
                for idx, coordinate in enumerate(thoracic_centerline[::-1]):
                    if coordinate[0] <= split:
                        arch_end = len(thoracic_centerline) - idx
                        break
                if arch_start is None or arch_end is None:
                    logger.error("Aortic arch could not be defined from morphology")
                    raise ValueError("Aortic arch could not be defined from morphology")
            # extract the aortic root as a separate region from the ascending aorta
            ascending_start = 0
            for idx, length in enumerate(self.centerline.cumulative_lengths):
                if length > self.root_length_mm:
                    ascending_start = idx
                    break
            return {
                "root": thoracic_indices[:ascending_start],
                "asc": thoracic_indices[ascending_start:arch_start],
                "arch": thoracic_indices[arch_start:arch_end],
                "desc": thoracic_indices[arch_end:],
            }

        if not self.centerline:
            raise ValueError(
                "Centerline was not found for the aorta. Cannot split aortic regions of the CPR."
            )
        elif not self.cpr.any():
            raise ValueError(
                "CPR is empty. Cannot split Aorta CPR into appropriate anatomical regions."
            )
        for region, info in self.regions.items():
            # If the region is not in the image, we move on
            if not info["in_image"]:
                continue
            if region == "thoracic":
                self.regions_by_index.update(_split_thoracic_aorta_by_index())
            else:
                self.regions_by_index[region] = (
                    self.centerline.get_indices_within_dimensional_coordinates(
                        info["start"], info["end"]
                    )
                )
        return self

    def aggregate_stats(self):
        """Aggregate all the desired stats into a dictionary to be sent out for storage"""
        aggregated = {}
        aggregated["whole"] = self._aggregate_region(include_diameters=False)
        for region, indices in self.regions_by_index.items():
            for index in indices:
                self.cross_section_stats[index].anatomical_region = region
            if len(indices) < 3:
                logger.warning(
                    f"Skipping aortic region {region} with less than 3 indices."
                )
                continue
            logger.debug(f"Aggregating stats for aortic region: {region}")
            # we only run these if the whole thing is there
            if region in ["root", "asc", "arch"]:
                entire_region = 1
            elif region == "desc" and "root" in self.regions_by_index:
                entire_region = 1
            else:
                entire_region = int(self.regions[region]["entire_region"])
            aggregated[region] = {
                "entire_region": entire_region,
                **self._aggregate_region(region_indices=indices),
            }
        # Denote if the entire aorta is present
        all_regions = {"whole", "root", "asc", "arch", "desc", "up_abd", "lw_abd"}
        if all(
            region.get("entire_region", 1) for region in aggregated.values()
        ) and all_regions == set(aggregated.keys()):
            aggregated["whole"]["entire_region"] = 1
        else:
            aggregated["whole"]["entire_region"] = 0
        self.stats = aggregated
        return self

    def _aggregate_region(
        self, region_indices: list[int] | None = None, include_diameters: bool = True
    ) -> dict:
        """
        Aggregate a specific region of the aorta using indices to extract from cross_section_stats

        Parameters:
            region_indices: list[int] - the list of indices to extract from the cross_section_stats. Defaults to entire aorta
        Returns:
            a dictionary containing the aggregated stats for the region
        """
        if region_indices is None:
            region_indices = [i for i in range(self.centerline.len())]
        # Get the region

        region = [self.cross_section_stats[x] for x in region_indices]
        start_idx = region_indices[0]
        end_idx = region_indices[-1]
        # Length is just the cumulative length over the region
        length = round(
            self.centerline.cumulative_lengths[end_idx]
            - self.centerline.cumulative_lengths[start_idx],
            1,
        )
        # Tortuosity from start to end
        tortuosity = self.centerline.calculate_tortuosity(start_idx, end_idx)
        # Initialize cumulative metrics
        ## Periaortic fat data
        peria_vol = 0
        peria_ring_vol = 0
        peria_fat_vol = 0
        peria_hus = []
        ## Calcification data
        calc_volume = 0
        calc_agatston = 0
        n_calcs = 0
        for cs in region:
            peria_vol += cs.peria_total_volume
            peria_ring_vol += cs.peria_ring_volume
            peria_fat_vol += cs.peria_fat_volume
            peria_hus.extend(cs.peria_fat_hus)
            calc_volume += cs.calc_volume_score
            calc_agatston += cs.calc_agatston_score
            n_calcs += cs.calc_regions
        peria_hu_mean, peria_hu_std = (
            np.array(peria_hus).mean(),
            np.array(peria_hus).std(),
        )
        if include_diameters:
            diameters = self._extract_diameters(region)
        else:
            diameters = {}

        return {
            "length_mm": length,
            **tortuosity,
            "peria_volume_cm3": round(peria_vol / 1000, 2),
            "peria_ring_volume_cm3": round(peria_ring_vol / 1000, 2),
            "peria_fat_volume_cm3": round(peria_fat_vol / 1000, 2),
            "peria_hu_mean": round(peria_hu_mean, 2),
            "peria_hu_std": round(peria_hu_std, 2),
            "calc_volume_mm3": round(calc_volume, 1),
            "calc_agatston": round(calc_agatston, 1),
            "calc_count": n_calcs,
            **diameters,
        }

    def _extract_diameters(
        self, region: list[AortaStraightenedCPRCrossSection]
    ) -> dict:
        diameters = {}
        # Filter out any cross section missing diameters or too small to be sure about
        diam_cross_sections = []
        for cs in region:
            if cs.mean_diameter is not None:
                if cs.mean_diameter > 5:
                    diam_cross_sections.append(cs)
        # if it's empty, return the empty dictionary
        if len(diam_cross_sections) == 0:
            return diameters
        # Metrics across the region
        start_idx = region[0].index
        end_idx = region[-1].index
        rel_idx = end_idx - start_idx
        mean_diameter = np.mean([x.mean_diameter for x in diam_cross_sections]).round(1)
        mean_roundness = np.mean([x.roundness for x in diam_cross_sections]).round(2)
        mean_flatness = np.mean([x.flatness for x in diam_cross_sections]).round(2)
        # Max diameter stats in the region
        max_diam = max(diam_cross_sections, key=lambda x: x.mean_diameter)
        mid_diam = region[len(region) // 2]  # This should basically never be missing
        if mid_diam.mean_diameter is None:
            mid_diam = diam_cross_sections[len(diam_cross_sections) // 2]
        min_diam = min(diam_cross_sections, key=lambda x: x.mean_diameter)
        # first measured cross section
        prox_diam = diam_cross_sections[0]
        # Last measured cross section
        dist_diam = diam_cross_sections[-1]
        rel_dist = lambda x: round((x.index - start_idx) / rel_idx, 2)
        rel_dists = {
            "max": rel_dist(max_diam),
            "min": rel_dist(min_diam),
            "prox": rel_dist(prox_diam),
            "mid": rel_dist(mid_diam),
            "dist": rel_dist(dist_diam),
        }
        return {
            "mean": {
                "diameter": mean_diameter,
                "roundness": mean_roundness,
                "flatness": mean_flatness,
            },
            "max": {
                "rel_distance": rel_dists["max"],
                "mean_diameter": max_diam.mean_diameter,
                "major_diameter": max_diam.major_diameter,
                "minor_diameter": max_diam.minor_diameter,
                "area": max_diam.area,
                "roundness": max_diam.roundness,
                "flatness": max_diam.flatness,
            },
            "min": {
                "rel_distance": rel_dists["min"],
                "mean_diameter": min_diam.mean_diameter,
                "major_diameter": min_diam.major_diameter,
                "minor_diameter": min_diam.minor_diameter,
                "area": min_diam.area,
                "roundness": min_diam.roundness,
                "flatness": min_diam.flatness,
            },
            "mid": {
                "rel_distance": rel_dists["mid"],
                "mean_diameter": mid_diam.mean_diameter,
                "major_diameter": mid_diam.major_diameter,
                "minor_diameter": mid_diam.minor_diameter,
                "area": mid_diam.area,
                "roundness": mid_diam.roundness,
                "flatness": mid_diam.flatness,
            },
            "proximal": {
                "rel_distance": rel_dists["prox"],
                "mean_diameter": prox_diam.mean_diameter,
                "major_diameter": prox_diam.major_diameter,
                "minor_diameter": prox_diam.minor_diameter,
                "area": prox_diam.area,
                "roundness": prox_diam.roundness,
                "flatness": prox_diam.flatness,
            },
            "distal": {
                "rel_distance": rel_dists["dist"],
                "mean_diameter": dist_diam.mean_diameter,
                "major_diameter": dist_diam.major_diameter,
                "minor_diameter": dist_diam.minor_diameter,
                "area": dist_diam.area,
                "roundness": dist_diam.roundness,
                "flatness": dist_diam.flatness,
            },
        }

    def calculate_stats(
        self, sampling_factor: float = 0.67, cross_section_radius: int = 100
    ):
        """
        Top-level method to calculate basic statistics for the aorta segmentation.
        Will create and process the centerline, create a straightened CPR,
        and calculate cross-section statistics.

        Parameters:
            sampling_factor: float - factor to downsample the centerline for smoothing and resampling - defaults to 0.67 (take 2 out of every 3 points).
            cross_section_radius: int - radius of the generated cross section in mm- defaults to 50.
        Returns:
            The Aorta instance with all statistics calculated stored as `cpr_cross_section_stats`.
        ### WILL EVENTUALLY BE DEPRECATED FOR A FULL AORTA STATISTICS CALCULATION METHOD
        return - The Aorta instance with all statistics calculated stored as `cpr_cross_section_stats`.
        """
        (
            self.determine_anatomical_regions()
            .create_centerline()
            .process_centerline(round(sampling_factor * self.centerline.len()))
            .create_straightened_cprs(cross_section_radius)
            .calculate_straightened_cpr_cross_section_stats()
            .split_anatomical_regions_by_indicies()
            .aggregate_stats()
        )
        return self

    def write_aorta_summary(self, output_path: str | Path, separator: str = ","):
        output_path = Path(output_path)
        col_order = {
            "index": "index",
            "anatomical_region": "region",
            "mean_diameter": "mean_diameter_mm",
            "major_diameter": "major_diameter_mm",
            "minor_diameter": "minor_diameter_mm",
            "area": "area",
            "roundness": "roundness",
            "flatness": "flatness",
            "peria_total_volume": "periaortic_total_volume_mm3",
            "peria_ring_volume": "periaortic_ring_volume_mm3",
            "peria_fat_volume": "periaortic_fat_volume_mm3",
            "calc_volume_score": "calc_volume_score",
            "calc_agatston_score": "calc_agatston_score",
            "calc_regions": "calc_regions",
        }
        data = []
        for cs in self.cross_section_stats:
            row = {col: getattr(cs, key) for key, col in col_order.items()}
            if row["region"] is None or row["mean_diameter_mm"] is None:
                continue
            for col, val in row.items():
                if isinstance(val, float):
                    row[col] = round(val, 2)
                elif isinstance(val, int):
                    row[col] = int(val)
            data.append(row)
        if output_path.suffix == ".gz":
            with gzip.open(output_path, mode="wt", newline="") as f:
                writer = csv.DictWriter(
                    f, fieldnames=col_order.values(), delimiter=separator
                )
                writer.writeheader()
                writer.writerows(data)
        else:
            with open(output_path, mode="w", newline="") as f:
                writer = csv.DictWriter(
                    f, fieldnames=col_order.values(), delimiter=separator
                )
                writer.writeheader()
                writer.writerows(data)



import torch
import SimpleITK as sitk

from multiprocessing import Pool
from datetime import date
from pathlib import Path
from loguru import logger
from nnunetv2.inference.predict_from_raw_data import nnUNetPredictor
from threadpoolctl import threadpool_limits

from mircat_v2.configs import read_models_config
from mircat_v2.dbase import insert_data_batch
from mircat_v2.nifti import SegNifti, resample_with_sitk


class MircatSegmentor:
    def __init__(
        self,
        niftis: list[Path],
        task_list: list[str],
        model_types: list[str],
        device: str,
        task_configs: dict | None = None,
        n_processes: int = 1,
        threads_per_process: int = 4,
        cache_size: int = 1,
        dbase_config: dict = {},
        img_resampler: str = "bspline",
        lbl_resampler: str = "gaussian",
    ):
        """Initialize segmentor tasks.
        :param task_list: list of all desired segmentation tasks
        :param model_types: list of task model types [3d, 2d, etc.]'
        :param device: the device for pytorch to use
        :param task_configs: dictionary containing task configs. If not passed, will load from library.
        :param n_processes: the number of workers for preprocessing, nnUNet, and postprocessing
        :param threads_per_process: the number of threads each worker can use.
        :param cache_size: Number of nifti files to work on at one time. Default = 1.
        :param dbase_config: dictionary containing database configuration. If empty, no database insertion will be performed.
        :param img_resampler: the type of resampling to use for images, default is "bspline"
        :param lbl_resampler: the type of resampling to use for labels, default is "gaussian"
        """
        self.niftis = niftis
        self.task_list = task_list
        self.model_types = model_types
        self.device = device
        self.n_processes = n_processes
        self.threads_per_process = threads_per_process
        self.cache_size = cache_size
        self.dbase_config = dbase_config
        self.dbase_insert = not dbase_config == {}
        self.img_resampler = img_resampler
        self.lbl_resampler = lbl_resampler
        # Logic for loading task configurations
        if isinstance(task_configs, dict):
            self.task_configs = task_configs
        elif task_configs is None:
            self.task_configs = read_models_config()
        else:
            message = "task_configs argument must be a dictionary or None."
            logger.error(message)
            raise ValueError(message)
        # Get the voxel spacings for each task/model type pair
        self.task_spacings = {
            task: tuple(self.task_configs[task]["models"][model]["spacing"])
            for task, model in zip(task_list, model_types)
        }

        self.task_model_paths = {
            task: self.task_configs[task]["models"][model]["path"]
            for task, model in zip(task_list, model_types)
        }

    def run(self):
        """Summary method to run all specified tasks on all given niftis
        :param niftis: path to either a single nifti file or a text file containing paths (preferably absolute paths) to all niftis for segmentation.
        """
        self._get_nifti_batches()
        for i, nifti_batch in enumerate(self.nifti_batches, start=1):
            self.current_batch = i
            self.run_batch(nifti_batch)

    def run_batch(self, nifti_batch: list[Path]):
        good_niftis, bad_niftis = self._load_nifti_batch(nifti_batch)
        self.niftis_to_segment = good_niftis
        self.failed_niftis = bad_niftis
        self._preprocess_batch()
        self._run_tasks()
        self._postprocess_batch()
        if self.dbase_insert:
            self._insert_batch_to_dbase()
        logger.success(
            "Batch [{}/{}] segmentation complete!",
            self.current_batch,
            self.total_batches,
        )
        # logic for uploading bad_nifti results to dbase

    def _load_nifti_batch(self, nifti_batch: list[Path]):
        """Load niftis in the batch and filter out bad ones
        :param nifti_batch: the list of nifti paths
        """
        good_niftis = []
        bad_niftis = []
        for nifti_path in nifti_batch:
            try:
                nifti = SegNifti(nifti_path)
                good_niftis.append(nifti)
            except Exception as e:
                bad_niftis.append(
                    {
                        "nifti": str(nifti_path),
                        "seg_date": date.today().isoformat(),
                        "status": "FAILED",
                        "failed_error": type(e).__name__,
                    }
                )
        logger.debug(
            f"Found {len(good_niftis)} niftis able to be loaded by SimpleITK for segmentation, failed to load {len(bad_niftis)} niftis."
        )
        return good_niftis, bad_niftis

    def _preprocess_batch(self) -> None:
        """Preprocess each nifti in a batch by writing each necessary temporary file
        :param good_niftis: the list of pre-validated Niftis instances from _load_nifti_batch
        """
        preprocessed = []
        logger.info(
            "Starting preprocessing for batch [{}/{}]",
            self.current_batch,
            self.total_batches,
        )
        with Pool(self.n_processes) as pool:
            for i, nifti in enumerate(
                pool.imap_unordered(self._preprocess_nifti, self.niftis_to_segment),
                start=1,
            ):
                preprocessed.append(nifti)
                logger.info(
                    "Batch [{}/{}]: [{}/{}] preprocessed",
                    self.current_batch,
                    self.total_batches,
                    i,
                    len(self.niftis_to_segment),
                )
        logger.success(
            "Batch [{}/{}] preprocessed!", self.current_batch, self.total_batches
        )
        self.niftis_to_segment = preprocessed

    def _preprocess_nifti(self, nifti: SegNifti) -> SegNifti:
        """Preprocess a single nifti file for segmentation by creating the necessary spacing files and storing the information
        :param nifti: An instance of the Nifti class
        """
        limits = threadpool_limits(self.threads_per_process)
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(self.threads_per_process)
        with limits:
            if not nifti.seg_folder.exists():
                logger.debug("Making segmentation folder {}", nifti.seg_folder)
                nifti.seg_folder.mkdir()
            else:
                logger.debug(
                    "Segmentation folder aready found for {}", nifti.seg_folder
                )
            preprocessed_files = {}
            for task, spacing in self.task_spacings.items():
                spacing_str: str = "_".join([str(x) for x in spacing])
                preprocessed_file: Path = (
                    nifti.seg_folder / f"{nifti.name}_{spacing_str}.nii.gz"
                )
                output_file: Path = (
                    nifti.seg_folder / f"{nifti.name}_{spacing_str}_{task}"
                )
                if not preprocessed_file.exists():
                    logger.debug(f"Making {preprocessed_file} for task {task}..")
                    nifti.resample_and_save_for_segmentation(
                        spacing, preprocessed_file, self.img_resampler
                    )
                else:
                    logger.debug(
                        f"Preprocessed file {preprocessed_file} already exists for task {task}."
                    )
                preprocessed_files[task] = {
                    "input": preprocessed_file,
                    "output": output_file,
                }
            nifti.task_files = preprocessed_files
            return nifti

    def _get_nifti_batches(self) -> None:
        """Ensure that the input niftis are in the correct format. If the input is a text file, we read the contents to get the paths of all niftis."""
        self.nifti_batches = [
            self.niftis[i : i + self.cache_size]
            for i in range(0, len(self.niftis), self.cache_size)
        ]
        self.total_batches = len(self.nifti_batches)
        logger.info(
            f"Running tasks {self.task_list} on {len(self.nifti_batches)} batches of {self.cache_size} niftis per batch."
        )

    def _run_tasks(self):
        device = torch.device(self.device)
        logger.info("Running on torch.device {}", device)
        for task in self.task_list:
            logger.info("Starting task {} with nnUNet", task)
            input_files = [
                [str(nifti.task_files.get(task).get("input"))]
                for nifti in self.niftis_to_segment
            ]
            output_files = [
                str(nifti.task_files.get(task).get("output"))
                for nifti in self.niftis_to_segment
            ]
            predictor = nnUNetPredictor(
                tile_step_size=0.5,
                use_gaussian=True,
                use_mirroring=False,
                perform_everything_on_device=True,
                device=device,
                verbose=False,
                verbose_preprocessing=False,
                allow_tqdm=True,
            )
            model_path = self.task_model_paths.get(task)
            if not isinstance(model_path, str) or not model_path:
                raise ValueError(
                    f"Model path for task '{task}' is missing or not a string."
                )
            predictor.initialize_from_trained_model_folder(
                model_path,
                use_folds=(0,),
                checkpoint_name="checkpoint_final.pth",
            )
            predictor.predict_from_files(
                input_files,
                output_files,
                save_probabilities=False,
                overwrite=True,
                num_processes_preprocessing=self.n_processes,
                num_processes_segmentation_export=self.n_processes,
            )
            del predictor

    def _postprocess_batch(self):
        n_niftis = len(self.niftis_to_segment)
        completed = []
        with Pool(self.n_processes) as pool:
            for i, dbase_insert in enumerate(
                pool.imap_unordered(self._postprocess_nifti, self.niftis_to_segment),
                start=1,
            ):
                completed.extend(dbase_insert)
                logger.info(
                    f"Batch [{self.current_batch}/{self.total_batches}]: {i}/{n_niftis} postprocessed"
                )
        self.completed_niftis = completed

    def _postprocess_nifti(self, nifti: SegNifti):
        sitk.ProcessObject.SetGlobalDefaultNumberOfThreads(self.threads_per_process)
        dbase_insert = []
        with threadpool_limits(self.threads_per_process):
            for task, task_files in nifti.task_files.items():
                input_nifti = task_files["input"]
                tmp_output_nifti = Path(f"{task_files['output']}.nii.gz")
                resampled_name = nifti.seg_folder / f"{nifti.name}_{task}.nii.gz"
                # Delete the preprocessed input
                if input_nifti.exists():
                    input_nifti.unlink()
                # Resample the output and create the new output file, delete the old after
                if tmp_output_nifti.exists():
                    lbl = sitk.ReadImage(str(tmp_output_nifti))
                    new_lbl = resample_with_sitk(
                        lbl,
                        new_size=nifti.shape,
                        is_label=True,
                        interpolator_type=self.lbl_resampler,
                    )
                    sitk.WriteImage(new_lbl, resampled_name)
                    logger.debug("successfully wrote {}", resampled_name)
                    tmp_output_nifti.unlink()
                else:
                    logger.warning(f"Output file {tmp_output_nifti} not found!")
                dbase_task = {
                    "nifti": str(nifti.path),
                    "series_uid": nifti.metadata.get("series_uid"),
                    "task": int(task),
                    "seg_file": str(resampled_name),
                    "seg_date": date.today().isoformat(),
                    "status": "SUCCESS",
                    "failed_error": None,
                }
                dbase_insert.append(dbase_task)
        return dbase_insert

    def _insert_batch_to_dbase(self):
        logger.info(
            "Batch [{}/{}]: Inserting results into mircat-v2 database",
            self.current_batch,
            self.total_batches,
        )
        insert_data_batch(self.dbase_config, "segmentations", self.completed_niftis)
        logger.debug(
            "Batch [{}/{}]: completed niftis successfully inserted",
            self.current_batch,
            self.total_batches,
        )
        insert_data_batch(self.dbase_config, "segmentations", self.failed_niftis)
        logger.debug(
            "Batch [{}/{}]: failed niftis successfully inserted",
            self.current_batch,
            self.total_batches,
        )
        logger.success(
            "Batch [{}/{}]: Results have been inserted into mircat-v2 database!",
            self.current_batch,
            self.total_batches,
        )



import numpy as np

from dataclasses import dataclass, field
from loguru import logger
from scipy.interpolate import splprep, splev


@dataclass
class Centerline:
    """
    Class representing a centerline for a vessel.
    """

    label: int
    vessel: str
    coordinates: np.ndarray
    _differences: np.ndarray = field(init=False, repr=False, default=None)
    _segment_lengths: np.ndarray = field(init=False, repr=False, default=None)
    _cumulative_lengths: np.ndarray = field(init=False, repr=False, default=None)
    _length: float = field(init=False, repr=False, default=None)
    u: np.ndarray = field(init=True, repr=False, default=None)
    tck: np.ndarray = field(init=True, repr=False, default=None)
    inflection_points: np.ndarray = field(init=True, repr=False, default=None)

    def __call__(self):
        """
        Return the centerline coordinates.
        """
        return self.coordinates

    def __len__(self):
        """
        Return the number of points in the centerline.
        """
        return len(self.coordinates)

    def len(self) -> int:
        return len(self.coordinates)

    @property
    def differences(self) -> float:
        if self._differences is None:
            self._differences = np.diff(self.coordinates, axis=0)
        return self._differences

    @property
    def segment_lengths(self) -> np.ndarray:
        if self._segment_lengths is None:
            self._segment_lengths = np.sqrt(np.sum(self.differences**2, axis=1))
        return self._segment_lengths

    @property
    def cumulative_lengths(self) -> np.ndarray:
        if self._cumulative_lengths is None:
            self._cumulative_lengths = np.concatenate(
                ([0], np.cumsum(self.segment_lengths))
            )
        return self._cumulative_lengths

    @property
    def length(self) -> float:
        # This is the length in voxel space - not in physical space. Need to adjust for spacing if physical length is desired.
        if self._length is None:
            self._length = self.cumulative_lengths[-1]
        return self._length

    def _reset_properties(self):
        """
        Reset all cached properties to force recalculation.
        """
        self._differences = None
        self._segment_lengths = None
        self._cumulative_lengths = None
        self._length = None

    def resample_with_spline(
        self, target_points: int, smoothing: float = 1.0, degree: int = 3
    ):
        """
        Resample centerline using B-spline interpolation for smoother results.

        Args:
            target_points: Number of points in resampled centerline
            smoothing: Smoothing factor (0 = interpolation, >0 = approximation)
            degree: Spline degree (1=linear, 2=quadratic, 3=cubic)
        """
        if target_points <= 0:
            raise ValueError("Target number of points must be greater than zero.")

        if len(self.coordinates) < degree + 1:
            raise ValueError(
                f"Need at least {degree + 1} points for degree {degree} spline"
            )

        # Prepare coordinates for splprep (transpose for per-dimension arrays)
        centerline = self.coordinates
        coords_t = centerline.T
        # Get the u parameter
        dists = np.linalg.norm(np.diff(centerline, axis=0), axis=1)
        u = np.concatenate(([0], np.cumsum(dists)))
        u /= u[-1]
        # Fit B-spline to the centerline
        # u is the parameter array, automatically computed based on chord length
        m = len(centerline)
        min_s = m - np.sqrt(2 * m)
        max_s = m + np.sqrt(2 * m)
        s = smoothing * len(centerline)
        if s < min_s:
            logger.warning("spline smoothing factor `s` is smaller than recommended!")
        elif s > max_s:
            logger.warning("spline smoothing factor `s` is greater than recommended!")
        tck, u = splprep(coords_t, u=u, s=s, k=degree)
        # Generate new parameter values for resampling
        u_new = np.linspace(0, 1, target_points)
        # Evaluate spline at new parameter values
        resampled_coords = splev(u_new, tck)
        resampled_coords = np.array(resampled_coords).T.round().astype(int)
        # Update the centerline coordinates
        self.coordinates = resampled_coords
        self.u = u_new
        self.tck = tck
        self._reset_properties()
        return self

    def calculate_tangent_vectors(self):
        """
        Calculate tangent vectors, normals, and binormals for the centerline.
        Returns:
            tangents: Tangent vectors at each point
            normals: Normal vectors at each point
            binormals: Binormal vectors at each point
        """
        if self.tck is None:
            raise ValueError(
                "Centerline must be resampled with a spline before calculating tangent vectors."
            )
        # Evaluate the spline to get the tangent vectors
        tangents = np.array(splev(self.u, self.tck, der=1)).T
        tangents /= np.linalg.norm(tangents, axis=1, keepdims=True)
        self.tangents = tangents
        return self

    def get_indices_within_dimensional_coordinates(
        self, start: int = None, end: int = None, dim: int = 0
    ) -> list[int]:
        """
        Method to extract a specific range of indices of a centerline from a start and end coordinate in a specific dimension.

        Parameters:
            start: int - the starting coordinate in the specified dimension (inclusive). If None - assumes the start of the dimension (0)
            end: int - the ending coordinate in the specified dimension (inclusive). If None - will be set to self.coordinates.max(dim) + 1
            dim: int - the dimension to match indices to the centerline is. Default is 0, which is z-dimension in mircat-v2
        Returns:
            a list containing the indices of the centerline that are within the range in the specified dimension.

        The start and end coordinates are ***inclusive*** in this method, example:
        If you had a coordinate in the centerline with the value of 5, and a start or end coordinate of 5, that point would be considered valid
        """
        if start is None:
            start = 0
        if end is None:
            end = self.coordinates.max(dim)
        assert start < end
        valid_indicies: list[int] = []
        for idx, coordinate in enumerate(self.coordinates):
            if start <= coordinate[dim] < end:
                valid_indicies.append(idx)
        return valid_indicies

    def calculate_tortuosity(self, start_idx=None, end_idx=None) -> dict:
        """
        Caclulate the totuosity index of the centerline, optionally from a specific start and end index

        Parameters:
            start_idx: The index of the centerline to start from (inclusive). Defaults to 0
            end_idx: The index of the centerline to end on (non-inclusive). Defaults to entire centerline
        Returns:
            A dictionary containing the measured tortuosity metrics.
        """
        if self.inflection_points is None:
            self.calculate_inflection_points()
        if start_idx is None:
            start_idx = 0
        if end_idx is None:
            end_idx = len(self.coordinates)
        tortuosity_coords = self.coordinates[start_idx:end_idx]
        inflection_points_in_range = np.array(
            [
                x
                for x in set(tuple(x) for x in tortuosity_coords)
                & set(tuple(x) for x in self.inflection_points)
            ]
        )
        # This the direct distance between the start and end coordinates
        # we don't have to worry about units because we will be dividing by the same unit
        euclidean_distance = np.linalg.norm(
            tortuosity_coords[-1] - tortuosity_coords[0]
        )
        # We shift the cumulative lengths by the length already traveled to the start index
        cumulative_lengths = (
            self.cumulative_lengths[start_idx:end_idx]
            - self.cumulative_lengths[start_idx]
        )[-1]
        # Calculate the two metrics
        tortuosity_index = round(cumulative_lengths / euclidean_distance, 3)
        icm = round((len(inflection_points_in_range) + 1) * tortuosity_index, 3)
        return {
            "tortuosity_index": tortuosity_index,
            "icm": icm,
            "n_inflections": len(inflection_points_in_range),
            # 'inflection_coords': inflection_points_in_range.tolist()
        }

    def calculate_inflection_points(self):
        """
        Internal method to calculate the coordinates of the inflection points along the centerline.

        An inflection point occurs where the curve changes from being concave to convex or vice versa.

        Returns:
            dict: A dictionary containing:
                - 'count': The number of inflection points
                - 'indices': List of indices of the inflection points
                - 'coordinates': List of coordinates of the inflection points
        """
        if self.tck is None:
            raise ValueError(
                "Centerline must be resampled with a spline before calculating inflection points."
            )

        # Calculate first and second derivatives
        first_derivatives = np.array(splev(self.u, self.tck, der=1)).T
        second_derivatives = np.array(splev(self.u, self.tck, der=2)).T

        # Normalize tangents
        tangents = first_derivatives / np.linalg.norm(
            first_derivatives, axis=1, keepdims=True
        )

        # Calculate curvature magnitudes
        curvature = np.zeros(len(self.u))
        for i in range(len(tangents)):
            # Remove tangential component from second derivative
            normal_component = (
                second_derivatives[i]
                - np.dot(second_derivatives[i], tangents[i]) * tangents[i]
            )
            curvature[i] = np.linalg.norm(normal_component)

        # Find local minima in the curvature
        local_min_indices = []
        for i in range(1, len(curvature) - 1):
            if curvature[i] < curvature[i - 1] and curvature[i] < curvature[i + 1]:
                local_min_indices.append(i)

        # Filter out minima that aren't close to zero (true inflection points have near-zero curvature)
        threshold = np.mean(curvature) * 0.3  # 30% of mean curvature as threshold
        inflection_indices = [
            idx for idx in local_min_indices if curvature[idx] < threshold
        ]

        # Get the coordinates of the inflection points
        inflection_points = (
            self.coordinates[inflection_indices] if inflection_indices else np.array([])
        )
        self.inflection_points = inflection_points



import SimpleITK as sitk
from loguru import logger
from mircat_v2.nifti import StatsNifti
from mircat_v2.stats.utilities import calculate_shape_stats, calculate_intensity_stats


def calculate_volume_and_intensity_stats(
    nifti: StatsNifti, seg_id: str, label_map: dict
) -> tuple[list[dict], dict]:
    """Calculate volume and intensity statistics for a given nifti file for the mircat-v2 custom segmentation.
    :param nifti: StatsNifti object containing the nifti file and its metadata.
    :param seg_id: The segmentation ID to calculate statistics for.
    :param label_map: Dictionary mapping structure names to label IDs.
    :return: Dictionary containing volume and intensity statistics.
    """
    logger.debug(f"Calculating volume and intensity stats for {nifti.name}.")
    segmentation = nifti.id_to_seg_map[seg_id]
    # Use SimpleITK to calculate the shape and intensity statistics
    shape_stats = calculate_shape_stats(segmentation)
    intensity_stats = calculate_intensity_stats(nifti.img, segmentation)
    # Get the average HU and volume for each label in the segmentation
    volumes_and_intensities = {}
    seg_labels = shape_stats.GetLabels()
    for structure, label in label_map.items():
        # Zero is background, so we skip it
        if label in seg_labels and label != 0:
            volume = round(shape_stats.GetPhysicalSize(label) / 1000, 1)
            avg_hu = round(intensity_stats.GetMean(label), 1)
            std_dev_hu = round(intensity_stats.GetStandardDeviation(label), 1)
            volumes_and_intensities[f"{structure}"] = {
                "volume_cm3": volume,
                "hu_mean": avg_hu,
                "hu_std_dev": std_dev_hu,
            }
    return volumes_and_intensities



from pathlib import Path
from math import floor
from multiprocessing import cpu_count
from loguru import logger

from mircat_v2.configs import set_threads_per_process
from mircat_v2.nifti import load_nifti_paths


def add_stats_subparser(subparsers):
    stats_parser = subparsers.add_parser(
        "stats",
        help="Calculate statistics on segmentation files",
        description="Calculate statistics on segmentation files in a directory.",
    )
    stats_parser.add_argument(
        "niftis",
        type=Path,
        help="Path to a nifti file with mircat-v2 segmentations or a text file with a list of multiple mircat-v2 nifti files.",
    )
    stats_parser.add_argument(
        "--data-dir",
        "-d",
        type=Path,
        default=Path(""),
        help="Data directory for the input niftis (i.e. {data_dir}/nifti_path) - useful for remapping in docker. Default is no data directory.",
    )
    stats_parser.add_argument(
        "--task-list",
        "-tl",
        type=str,
        nargs="+",
        default=["all"],
        choices=["all", "vol-int", "contrast", "vertebrae", "aorta", "tissues"],
        help="List of tasks to calculate statistics for. Default is 'all'. vol-int: volume and intensity statistics, contrast: contrast phase prediction, aorta: aorta specific statistics, tissues: tissue segmentation statistics.",
    )
    stats_parser.add_argument(
        "--resolution",
        "-r",
        type=str,
        choices=["normal", "high", "highest"],
        default="normal",
        help="Isotropic Resolution to transpose images and labels to for statistics. 'normal' is 1mm, 'high' is 0.75mm, and 'highest' is 0.5mm. Default is 'normal'.",
    )
    stats_parser.add_argument(
        "--image-resampler",
        "-ir",
        type=str,
        choices=["lanczos", "bspline", "gaussian", "linear"],
        default="bspline",
        help="Image resampling method to use for statistics. 'bspline' is the default and strikes the best balance between speed and quality. 'lanczos' is the slowest, but has the best quality. 'gaussian' is faster and good for most cases, 'linear' is the fastest but may not be suitable for all images.",
    )
    stats_parser.add_argument(
        "--label-resampler",
        "-lr",
        type=str,
        choices=["gaussian", "linear", "nearest"],
        default="gaussian",
        help="Label resampling method to use for statistics. 'gaussian' is the default and slowest, 'linear' is faster but has lower resolution, 'nearest' is the fastest but may not be suitable for all labels.",
    )
    stats_parser.add_argument(
        "--n-processes",
        "-n",
        type=int,
        default=1,
        help="Number of processes to use for parallel processing. Default is 1 (only one process).",
    )
    stats_parser.add_argument(
        "--threads-per-process",
        "-t",
        type=int,
        default=4,
        help="Number of threads per process for multithreaded processing for python libraries. Default is 4.",
    )
    stats_parser.add_argument(
        "--dbase-insert",
        "-db",
        action="store_true",
        help="Insert the calculated statistics into the database. Requires a database to be set up with mircat-v2 dbase create command. Default is False (do not insert into database).",
    )
    stats_parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Overwrite existing stats files if they exist. Default is False (do not overwrite).",
    )
    stats_parser.add_argument(
        "--gzip",
        action="store_true",
        help="Compress the output json stats file with gzip. Default is False (do not compress).",
    )


def run_stats(args):
    """Run the statistics calculation on the provided NiFTi files."""
    from mircat_v2.stats.analyzer import Analyzer

    # Set up threads per process
    set_threads_per_process(args)
    logger.debug("Input arguments: {}", args)
    niftis = load_nifti_paths(args.niftis, args.data_dir)
    analyzer = Analyzer(
        niftis=niftis,
        task_list=args.task_list,
        resolution=args.resolution,
        image_resampler=args.image_resampler,
        label_resampler=args.label_resampler,
        n_processes=args.n_processes,
        threads_per_process=args.threads_per_process,
        dbase_insert=args.dbase_insert,
        overwrite=args.overwrite,
        gzip=args.gzip,
    )
    analyzer.run()



import json
import sys
from math import floor
from multiprocessing import cpu_count
from pathlib import Path
from loguru import logger

__all__ = [
    "config_path",
    "config_file",
    "dbase_schema_file",
    "read_config",
    "write_config",
    "read_segmentation_config",
    "read_dbase_config",
    "read_segmentation_config",
    "read_models_config",
    "read_stats_models_config",
    "logger_setup",
    "set_threads_per_process",
]


def generate_config() -> None:
    """Generate the default config"""
    with config_file.open("w") as f:
        json.dump(default_config, f, indent=2)


def read_config() -> dict[str, dict]:
    "Read and return the entire config file"
    try:
        with config_file.open() as f:
            return json.load(f)
    except json.decoder.JSONDecodeError as e:
        logger.error("the config.json file could not be decoded.")
        raise e


def write_config(new_config: dict, key=str, subkey: str = None) -> None:
    """Write a new configuration to a specific key within the config.json file
    Parameters:
        new_config: dict - the new configuration to write
        key: str - the key that the configuration should be stored in, inside the config.json
    """
    config = read_config()
    if subkey is not None:
        config[key][subkey] = new_config
    else:
        config[key] = new_config
    with config_file.open("w") as f:
        json.dump(config, f, indent=2)
    logger.success(
        f"{key}{'[' + subkey + ']' if subkey is not None else ''} configuration saved to {config_file}"
    )


def read_segmentation_config() -> dict[str, dict]:
    "Read the config file and return the segmentation configuration"
    with config_file.open() as f:
        return json.load(f).get("segmentation", {})


def read_models_config() -> dict[str, dict]:
    "Read the config file and return the segmentation model configuration"
    with config_file.open() as f:
        return json.load(f).get("models", {})


def read_stats_models_config() -> dict[str, dict]:
    "Read the config file and return the stats models configuration"
    with config_file.open() as f:
        return json.load(f).get("stats_models", {})


def read_dbase_config() -> dict[str, dict]:
    """Read the config file and return the database configuration"""
    with config_file.open() as f:
        return json.load(f).get("dbase", {})


def logger_setup(verbose: bool, quiet: bool) -> None:
    """Set up logger for mircat-v2
    :param verbose: be verbose in the output by adding debug to stdout
    :param quiet: be quiet in the output by only showing successes and errors - no warnings or info.
    """
    logger.remove()
    # Regular
    stdout_fmt = "<green>{time: DD-MM-YYYY -> HH:mm:ss}</green> | <level>{level}</level> | <level>{message}</level>"
    stderr_fmt = "<red>{time: DD-MM-YYYY -> HH:mm:ss}</red> | <level>{level}</level> | <level>{message}</level>"
    if quiet:
        # Only show success messages and error messages
        logger.add(
            sys.stdout,
            format=stdout_fmt,
            level="SUCCESS",
            filter=lambda record: record["level"].no <= 25,
            enqueue=True,
        )
        logger.add(sys.stderr, format=stderr_fmt, level="ERROR", enqueue=True)
    elif verbose:
        # Include debugging output
        logger.add(
            sys.stdout,
            format=stdout_fmt,
            level="DEBUG",
            filter=lambda record: record["level"].no <= 25,
            enqueue=True,
        )
        logger.add(
            sys.stderr,
            format=stderr_fmt,
            level="WARNING",
            enqueue=True,
        )
    else:
        # Show everything above INFO
        logger.add(
            sys.stdout,
            format=stdout_fmt,
            level="INFO",
            filter=lambda record: record["level"].no <= 25,
            enqueue=True,
        )
        logger.add(
            sys.stderr,
            format=stderr_fmt,
            level="WARNING",
            enqueue=True,
        )


def set_threads_per_process(args):
    # Do some thread matching
    total_threads = cpu_count()
    required_threads = args.n_processes * args.threads_per_process
    logger.debug(
        f"Total threads on machine: {total_threads}. Requested threads for segmentation: {required_threads}"
    )
    if total_threads < required_threads:
        args.threads_per_process = floor(total_threads / args.n_processes)
        logger.warning(
            "Desired threads (n_processes * threads_per_process) > total threads on device ({}>{}). Limiting threads per worker to {}. If performance drops, consider reducing the number of workers and upping threads to match.",
            required_threads,
            total_threads,
            args.threads_per_process,
        )


config_path = Path(__file__).parent.resolve()
# Config will always be stored in the package files - easier to manage
config_file = config_path / "config.json"
dbase_schema_file = config_path / "dbase.sql"
db_schema = {
    "conversions": {
        "series_uid": {"type": "TEXT", "pk": 1},
        "study_uid": {"type": "TEXT", "pk": 0},
        "nifti": {"type": "TEXT", "pk": 0},
        "modality": {"type": "TEXT", "pk": 0},
        "mrn": {"type": "TEXT", "pk": 0},
        "accession": {"type": "TEXT", "pk": 0},
        "series_name": {"type": "TEXT", "pk": 0},
        "series_number": {"type": "INTEGER", "pk": 0},
        "scan_date": {"type": "TEXT", "pk": 0},
        "original_series_name": {"type": "TEXT", "pk": 0},
        "study_description": {"type": "TEXT", "pk": 0},
        "ct_direction": {"type": "TEXT", "pk": 0},
        "image_type": {"type": "TEXT", "pk": 0},
        "sex": {"type": "TEXT", "pk": 0},
        "age": {"type": "INTEGER", "pk": 0},
        "birth_date": {"type": "TEXT", "pk": 0},
        "height_m": {"type": "REAL", "pk": 0},
        "weight_kg": {"type": "REAL", "pk": 0},
        "pregnancy_status": {"type": "INTEGER", "pk": 0},
        "pixel_length_mm": {"type": "REAL", "pk": 0},
        "pixel_width_mm": {"type": "REAL", "pk": 0},
        "slice_thickness_mm": {"type": "REAL", "pk": 0},
        "manufacturer": {"type": "TEXT", "pk": 0},
        "model": {"type": "TEXT", "pk": 0},
        "kvp": {"type": "REAL", "pk": 0},
        "sequence_name": {"type": "TEXT", "pk": 0},
        "protocol_name": {"type": "TEXT", "pk": 0},
        "contrast_bolus_agent": {"type": "TEXT", "pk": 0},
        "contrast_bolus_route": {"type": "TEXT", "pk": 0},
        "contrast_bolus_volume": {"type": "REAL", "pk": 0},
        "dicom_folder": {"type": "TEXT", "pk": 0},
        "conversion_date": {"type": "TEXT", "pk": 0},
    },
    "segmentations": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "task": {"type": "INTEGER", "pk": 2},
        "seg_file": {"type": "TEXT", "pk": 0},
        "seg_date": {"type": "TEXT", "pk": 3},
        "status": {"type": "TEXT", "pk": 0},
        "failed_error": {"type": "TEXT", "pk": 0},
    },
    "metadata": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "study_uid": {"type": "TEXT", "pk": 0},
        "output_stats_file": {"type": "TEXT", "pk": 0},
        "modality": {"type": "TEXT", "pk": 0},
        "mrn": {"type": "TEXT", "pk": 0},
        "accession": {"type": "TEXT", "pk": 0},
        "series_name": {"type": "TEXT", "pk": 0},
        "series_number": {"type": "INTEGER", "pk": 0},
        "scan_date": {"type": "TEXT", "pk": 0},
        "original_series_name": {"type": "TEXT", "pk": 0},
        "study_description": {"type": "TEXT", "pk": 0},
        "ct_direction": {"type": "TEXT", "pk": 0},
        "abdominal_scan": {"type": "INTEGER", "pk": 0},
        "chest_scan": {"type": "INTEGER", "pk": 0},
        "correct_vertebrae_order": {"type": "INTEGER", "pk": 0},
        "lowest_vertebra": {"type": "TEXT", "pk": 0},
        "highest_vertebra": {"type": "TEXT", "pk": 0},
        "image_type": {"type": "TEXT", "pk": 0},
        "sex": {"type": "TEXT", "pk": 0},
        "age": {"type": "INTEGER", "pk": 0},
        "birth_date": {"type": "TEXT", "pk": 0},
        "height_m": {"type": "REAL", "pk": 0},
        "weight_kg": {"type": "REAL", "pk": 0},
        "pregnancy_status": {"type": "INTEGER", "pk": 0},
        "pixel_length_mm": {"type": "REAL", "pk": 0},
        "pixel_width_mm": {"type": "REAL", "pk": 0},
        "slice_thickness_mm": {"type": "REAL", "pk": 0},
        "manufacturer": {"type": "TEXT", "pk": 0},
        "model": {"type": "TEXT", "pk": 0},
        "kvp": {"type": "REAL", "pk": 0},
        "sequence_name": {"type": "TEXT", "pk": 0},
        "protocol_name": {"type": "TEXT", "pk": 0},
        "contrast_bolus_agent": {"type": "TEXT", "pk": 0},
        "contrast_bolus_route": {"type": "TEXT", "pk": 0},
        "contrast_bolus_volume": {"type": "REAL", "pk": 0},
        "dicom_folder": {"type": "TEXT", "pk": 0},
        "conversion_date": {"type": "TEXT", "pk": 0},
    },
    "vol_int": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "structure": {"type": "TEXT", "pk": 2},
        "volume_cm3": {"type": "REAL", "pk": 0},
        "hu_mean": {"type": "REAL", "pk": 0},
        "hu_std_dev": {"type": "REAL", "pk": 0},
    },
    "contrast": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "phase": {"type": "TEXT", "pk": 0},
        "probability": {"type": "REAL", "pk": 0},
        "pi_time": {"type": "REAL", "pk": 0},
        "pi_time_std": {"type": "REAL", "pk": 0},
    },
    "vertebrae": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "vertebra": {"type": "TEXT", "pk": 2},
        "midline": {"type": "INTEGER", "pk": 0},
    },
    "aorta_metrics": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "region": {"type": "TEXT", "pk": 2},
        "entire_region": {"type": "INTEGER", "pk": 0},
        "length_mm": {"type": "REAL", "pk": 0},
        "tortuosity_index": {"type": "REAL", "pk": 0},
        "icm": {"type": "REAL", "pk": 0},
        "n_inflections": {"type": "INTEGER", "pk": 0},
        "peria_volume_cm3": {"type": "REAL", "pk": 0},
        "peria_ring_volume_cm3": {"type": "REAL", "pk": 0},
        "peria_fat_volume_cm3": {"type": "REAL", "pk": 0},
        "peria_hu_mean": {"type": "REAL", "pk": 0},
        "peria_hu_std": {"type": "REAL", "pk": 0},
        "calc_volume_mm3": {"type": "REAL", "pk": 0},
        "calc_agatston": {"type": "REAL", "pk": 0},
        "calc_count": {"type": "INTEGER", "pk": 0},
        "is_120_kvp": {"type": "INTEGER", "pk": 0},
        "mean_diameter_mm": {"type": "REAL", "pk": 0},
        "mean_roundness": {"type": "REAL", "pk": 0},
        "mean_flatness": {"type": "REAL", "pk": 0},
    },
    "aorta_diameters": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "region": {"type": "TEXT", "pk": 2},
        "measure": {"type": "TEXT", "pk": 3},
        "mean_diameter_mm": {"type": "REAL", "pk": 0},
        "major_diameter_mm": {"type": "REAL", "pk": 0},
        "minor_diameter_mm": {"type": "REAL", "pk": 0},
        "area_mm2": {"type": "REAL", "pk": 0},
        "roundness": {"type": "REAL", "pk": 0},
        "flatness": {"type": "REAL", "pk": 0},
        "rel_distance": {"type": "REAL", "pk": 0},
        "entire_region": {"type": "INTEGER", "pk": 0},
    },
    "tissues_volumetric": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "region": {"type": "TEXT", "pk": 2},
        "structure": {"type": "TEXT", "pk": 3},
        "volume_cm3": {"type": "REAL", "pk": 0},
        "hu_mean": {"type": "REAL", "pk": 0},
        "hu_std_dev": {"type": "REAL", "pk": 0},
    },
    "tissues_vertebral": {
        "nifti": {"type": "TEXT", "pk": 1},
        "series_uid": {"type": "TEXT", "pk": 0},
        "vertebra": {"type": "TEXT", "pk": 2},
        "structure": {"type": "TEXT", "pk": 3},
        "measurement": {"type": "TEXT", "pk": 4},
        "value": {"type": "REAL", "pk": 0},
    },
}

needed_stats = {
    "999": {
        "background": 0,
        "spleen": 1,
        "kidney_right": 2,
        "kidney_left": 3,
        "gallbladder": 4,
        "liver": 5,
        "stomach": 6,
        "pancreas": 7,
        "adrenal_gland_right": 8,
        "adrenal_gland_left": 9,
        "lung_upper_lobe_left": 10,
        "lung_lower_lobe_left": 11,
        "lung_upper_lobe_right": 12,
        "lung_middle_lobe_right": 13,
        "lung_lower_lobe_right": 14,
        "esophagus": 15,
        "trachea": 16,
        "thyroid_gland": 17,
        "small_bowel": 18,
        "duodenum": 19,
        "colon": 20,
        "urinary_bladder": 21,
        "prostate": 22,
        "sacrum": 23,
        "vertebrae_S1": 24,
        "vertebrae_L5": 25,
        "vertebrae_L4": 26,
        "vertebrae_L3": 27,
        "vertebrae_L2": 28,
        "vertebrae_L1": 29,
        "vertebrae_T12": 30,
        "vertebrae_T11": 31,
        "vertebrae_T10": 32,
        "vertebrae_T9": 33,
        "vertebrae_T8": 34,
        "vertebrae_T7": 35,
        "vertebrae_T6": 36,
        "vertebrae_T5": 37,
        "vertebrae_T4": 38,
        "vertebrae_T3": 39,
        "vertebrae_T2": 40,
        "vertebrae_T1": 41,
        "vertebrae_C7": 42,
        "vertebrae_C6": 43,
        "vertebrae_C5": 44,
        "vertebrae_C4": 45,
        "vertebrae_C3": 46,
        "vertebrae_C2": 47,
        "vertebrae_C1": 48,
        "heart": 49,
        "aorta": 50,
        "pulmonary_vein": 51,
        "brachiocephalic_trunk": 52,
        "subclavian_artery_right": 53,
        "subclavian_artery_left": 54,
        "common_carotid_artery_right": 55,
        "common_carotid_artery_left": 56,
        "brachiocephalic_vein_left": 57,
        "brachiocephalic_vein_right": 58,
        "atrial_appendage_left": 59,
        "superior_vena_cava": 60,
        "inferior_vena_cava": 61,
        "portal_vein_and_splenic_vein": 62,
        "iliac_artery_left": 63,
        "iliac_artery_right": 64,
        "iliac_vena_left": 65,
        "iliac_vena_right": 66,
    },
    "485": {
        "background": 0,
        "subq_fat": 1,
        "visc_fat": 2,
        "skeletal_muscle": 3,
        "intermuscular_fat": 4,
    },
    "481": {"background": 0, "subq_fat": 1, "visc_fat": 2, "skeletal_muscle": 3},
    "299": {"background": 0, "body": 1, "body_extremities": 2},
    "300": {"background": 0, "body": 1, "body_extremities": 2},
}

# We want the standard dbase schema to always be there - all we need to update is the path to the database file
default_config = {
    "dbase": {"tables": db_schema},
    "stats_models": needed_stats,
    "models": {},
}
if not config_file.exists():
    generate_config()
